"""
    SSDC.MappingDecomposer.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from scipy.stats import linregress
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.ModelUtils import compute_cy_list, compute_area_props
from molass_legacy.Models.Stochastic.MonoporeMoments import study_monopore_moments_impl
from SecTheory.BasicModels import robust_single_pore_pdf

def guess_init_mapping(sorted_recs_list, moments_list):
    x = []
    y = []
    for (i, p), (j, q) in zip(sorted_recs_list[0][0:2], sorted_recs_list[1][0:2]):
        x.append(moments_list[0][i][0])
        y.append(moments_list[1][j][0])

    slope, intercept, r_value, p_value, std_err = linregress(x, y)
    return slope, intercept

def guess_mapping(source, target, props_list, moments_list):
    sorted_recs_list = []
    for props in props_list:
        sorted_recs = sorted([(i, p) for i, p in enumerate(props)], key=lambda x: -x[1])
        sorted_recs_list.append(sorted_recs)
    model = source.model
    sx = source.xr_x
    sy = source.xr_y
    tx = target.xr_x
    ty = target.xr_y

    sorted_recs = sorted_recs_list[0]
    if len(sorted_recs) < 2:        
        """
        tx = a*sx + b
        b = tx - a*sx
        """
        a = 1
        i = sorted_recs[0][0]
        b = moments_list[1][i][0] - a*moments_list[0][i][0]     # computing b using the most significant peak
        print("b=", b)
        mapping = a, b
    else:
        mapping = guess_init_mapping(sorted_recs_list, moments_list)
        return mapping

    speaks = source.get_peaks()
    tpeaks = target.get_peaks()
    src_params_list = []
    ret_params_list = []
    ret_indeces = []
    sry_ = sy.copy()
    try_ = ty.copy()
    for k, (i, p) in enumerate(sorted_recs, start=1):
        sparams = speaks[i]
        print([i], "sparams=", sparams)
        print([i], "tparams",  tpeaks[i])
        scale = np.max(try_)/np.max(sry_)
        ret = get_a_mapped_component_params(mapping, scale, model, sparams, sx, sry_, tx, try_, k, i, ret_params_list)
        if ret is None:
            return
        src_params_list.append(sparams)
        ret_params_list.append(ret)
        ret_indeces.append(i)
        mapping = update_mapping(mapping, src_params_list, ret_params_list)
        if k >= 3:
            break
        cy = model(sx, sparams)
        sry_ -= cy
        cy = model(tx, ret)
        try_ -= cy

    return mapping

def update_mapping(mapping, src_params_list, ret_params_list):
    a, b = mapping
    points = []
    if len(src_params_list) < 2:
        points.append((1, a+b))
    for sparams, tparams in zip(src_params_list, ret_params_list):
        points.append((sparams[1], tparams[1]))
    points_array = np.array(points)
    slope, intercept, r_value, p_value, std_err = linregress(points_array[:,0], points_array[:, 1])
    return slope, intercept

def compute_monopore_curves(x, mnp_params, rgs):
    N, T, t0, me, mp, poresize = mnp_params[0:6]
    scales = mnp_params[6:]
    rhov = rgs/poresize
    rhov[rhov > 1] = 1
    cy_list = []
    for k, (rho, scale) in enumerate(zip(rhov, scales)):
        ni_ = N * (1 - rho)**me
        ti_ = T * (1 - rho)**mp
        cy = scale * robust_single_pore_pdf(x - t0, ni_, ti_)
        cy_list.append(cy)
    tty = np.sum(cy_list, axis=0)
    return cy_list, tty

def decompose_paired_elution_curves(info_list, moments_list=None, num_components=None):
    print("decompose_paired_elution_curves")
    if moments_list is None:
        moments_list = []
        for in_folder, lrf_src in info_list:
            moments_list.append(lrf_src.get_egh_moments_list())
    
    props_list = []
    for in_folder, lrf_src in info_list:
        peaks = lrf_src.get_peaks()
        cy_list = compute_cy_list(lrf_src.model, lrf_src.xr_x, peaks)
        props = compute_area_props(cy_list)
        props_list.append(props)

    source = info_list[0][1]
    target = info_list[1][1]    
    # props = props_list[0]

    mapping = guess_mapping(source, target, props_list, moments_list)

    # sorted_recs = sorted([(i, p) for i, p in enumerate(props)], key=lambda x: -x[1])
    # print("sorted_recs=", sorted_recs)

    model = source.model
    sx = source.xr_x
    sy = source.xr_y
    tx = target.xr_x
    ty = target.xr_y

    # 
    ret_params, ret_rgs, unreliable_indeces, params_scaler = study_monopore_moments_impl(source, debug=False)
    in_folder = info_list[0][0]
    if in_folder.find("20210727") >= 0:
        unreliable_indeces = np.array([3])      # temporary data dependent fix
    elif in_folder.find("20220716/BSA_") >= 0:
            unreliable_indeces = np.array([0])  # temporary data dependent fix
    elif in_folder.find("20230706") >= 0:
            unreliable_indeces = np.array([0])  # temporary data dependent fix

    reliable_indeces = np.setdiff1d(np.arange(len(ret_rgs)), unreliable_indeces)
    print("reliable_indeces=", reliable_indeces)
    ret_scales = ret_params[6:]
    mnp_params = np.concatenate([ret_params[0:6], ret_scales[reliable_indeces]])
    temp_rgs = ret_rgs[reliable_indeces]

    cy_list, tty = compute_monopore_curves(sx, mnp_params, temp_rgs)

    a, b = mapping
    tspline = UnivariateSpline(tx, ty, s=0, ext=3)
    tx_ = a*sx + b
    ty_ = tspline(tx_)
    init_scale = np.max(ty)/np.max(sy)
    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
        fig.suptitle("Monopore Curve Mapping", fontsize=20)
        ax1.plot(sx, sy)
        ax2.plot(tx_, ty_)
        for k, cy in enumerate(cy_list):
            ax1.plot(sx, cy, ":", label='component-%d' % k)
            ax2.plot(tx_, init_scale*cy, ":", label='component-%d' % k)
        fig.tight_layout()
        ret = plt.show()
    if not ret:
        return

    n = len(mnp_params)
    num_peaks = len(temp_rgs)
    print("temp_rgs=", temp_rgs)
    print("num_peaks=", num_peaks)
    print("init mapping=", mapping)
    init_a, init_b = mapping

    def split_params(p):
        mnp_params_ = p[0:n]
        rgs_ = p[n:n+num_peaks] 
        a, b, scale = p[n+num_peaks:]
        return mnp_params_, rgs_, a, b, scale

    def mapping_objective(p):
        mnp_params_, rgs_, a, b, scale = split_params(p)
        scy_list, sty = compute_monopore_curves(sx, mnp_params_, rgs_)            
        tx_ = a*sx + b
        ty_ = tspline(tx_)
        mapping_penalty = max(0, abs(1 - a/init_a) - 0.05)**2 *1e6
        return np.log(np.sum((sty - sy)**2)) +  np.log(np.sum((scale*sty - ty_)**2)) + mapping_penalty

    init_sacle = np.max(ty)/np.max(sy)
    init_params = np.concatenate((mnp_params, temp_rgs, mapping, [init_sacle]))
    res = minimize(mapping_objective, init_params, method='Nelder-Mead')
    mnp_params_, rgs_, a, b, scale = split_params(res.x)
    print("optimized mapping=", (a, b))

    return mnp_params_, rgs_, (a, b), scale

def get_a_mapped_component_params(mapping, scale, model, sparams, sx, sy, tx, ty, k, i, ret_params_list, debug=False):
    a, b = mapping

    def plot_component_state(sparams, tparams, title=None):
        if title is None:
            title = "Component %d" % k

        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
            fig.suptitle(title)
            ax1.set_title("Source Decomposition State")
            ax1.plot(sx, sy, color="orange")
            cy = model(sx, sparams)
            ax1.plot(sx, cy, ":", color="C%d" % i)
            ax2.set_title("Target Decomposition State")
            ax2.plot(tx, ty, color="orange")
            cy = model(tx, tparams)
            ax2.plot(tx, cy, ":", color="C%d" % i)
            fig.tight_layout()
            ret = plt.show()
        return ret
    
    tparams = sparams.copy()
    tparams[0] = sparams[0]*scale
    tparams[1] = a*sparams[1] + b

    if debug:
        ret = plot_component_state(sparams, tparams, title="[%d-%d] befor optimize" % (k,i))
        if not ret:
            return

    m, s = sparams[1:3]
    target_range = []
    for px in m - 2*s, m + 2*s:
        target_range.append(a*px + b)
    
    start = max(0, int(target_range[0])) 
    stop = min(len(tx), int(target_range[1]))
    tx_ = tx[start:stop]
    ty_ = ty[start:stop]

    def objective(p):
        tparams[[0, 1]] = p
        return np.sum((ty_ - model(tx_, tparams))**2)

    res = minimize(objective, tparams[[0,1]])
    tparams[[0,1]] = res.x

    if debug:    
        ret = plot_component_state(sparams, tparams, title="[%d-%d] after optimize" % (k,i))
        if not ret:
            return

    return tparams



"""
    SecTheory.Cavazzini1999.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np

def single_site_cf(w, n, t):
    return np.exp(n*(1/(1 - 1j*w*t) - 1))

def two_site_cf(w, n, p1, t1, t2):
    p2 = 1 - p1
    return np.exp(n*(p1/(1 - 1j*w*t1) + p2/(1 - 1j*w*t2) - 1))

def three_site_cf(w, n, p1, p2, t1, t2, t3):
    p3 = 1 - p1 - p2
    return np.exp(n*(-1 + p1/(1 - 1j*w*t1) + p2/(1 - 1j*w*t2) + p3/(1 - 1j*w*t3)))

def uniform_adsorption_energy(w, n, t1, t2):
    return np.exp(n*(np.log((1 - 1j*w*t1)/(1 - 1j*w*t2))/np.log(t2/t1)))

def uniform_mean_sojourn_time(w, n, t1, t2):
    return np.exp(n*(1/(1j*w*(t2 - t1))*np.log((1 - 1j*w*t1)/(1 - 1j*w*t2)) - 1))



# basin hopping global optimization for the ackley multimodal objective function
from scipy.optimize import basinhopping
import numpy as np
from numpy.random import rand
from numpy import exp
from numpy import sqrt
from numpy import cos
from numpy import e
from numpy import pi
from numpy import meshgrid
from numpy import arange
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon
from matplotlib import animation

# ---- Take Step Cumtomization Begin -------------------------------------------
from scipy._lib._util import check_random_state
from scipy.optimize._basinhopping import RandomDisplacement, AdaptiveStepsize

takestep_history = []
class StepRecord:
    def __init__(self, x, ret_x, stepsize, adjusted):
        self.x = x
        self.ret_x = ret_x
        self.stepsize = stepsize
        self.adjusted = adjusted

    def get_vertices(self):
        x, y = self.x
        s = self.stepsize
        return [(x-s, y-s), (x+s, y-s), (x+s, y+s), (x-s, y+s)]

    def create_patch(self):
        vertices = self.get_vertices()
        return Polygon(vertices, alpha=0.3, color='green', label="hopping window")

    def move_patch(self, patch):
        vertices = self.get_vertices()
        patch.set_xy(vertices)

    def create_displine(self, ax):
        line, = ax.plot(*np.array([self.x, self.ret_x]).T, color="yellow", label="hopping")
        point, = ax.plot(*self.ret_x, "o", color="orange", label="after hopping")
        return line, point

    def move_displine(self, displace):
        line, point = displace
        xy = np.array([self.x, self.ret_x]).T
        line.set_data(*xy)
        x, y = self.ret_x
        point.set_data([x], [y])

class CustomTakestep(AdaptiveStepsize):
    def __init__(self, stepsize=0.5, interval=50, disp=False, seed=None):
        # set up the np.random generator
        self.rng = check_random_state(seed)
        displace = RandomDisplacement(stepsize=stepsize, random_gen=self.rng)
        AdaptiveStepsize.__init__(self, displace, interval=interval, verbose=disp)

    def take_step(self, x):
        self.nstep += 1
        self.nstep_tot += 1
        adjusted = False
        if self.nstep % self.interval == 0:
            self._adjust_step_size()
            adjusted = True
        x_ = np.copy(x)
        ret_x = self.takestep(x_)
        takestep_history.append(StepRecord(x, ret_x, self.takestep.stepsize, adjusted))
        return ret_x

# ---- Take Step Cumtomization End ---------------------------------------------

# objective function
def objective(v):
    x, y = v
    return -20.0 * exp(-0.2 * sqrt(0.5 * (x**2 + y**2))) - exp(0.5 * (cos(2 * pi * x) + cos(2 * pi * y))) + e + 20

# define range for input
r_min, r_max = -5.0, 5.0
# define the starting point as a random sample from the domain

minima_list = []
init_points = [(-4, -4), (-4, -4), (4, -4), (4, -4), (4, 4), (4, 4), (-4, 4), (-4, -4)]

def solve(k):

    stepsize = 0.5
    take_step = CustomTakestep(stepsize=0.5)

    # pt = r_min + rand(2) * (r_max - r_min)
    pt = init_points[k]

    takestep_history.append(StepRecord(pt, pt, stepsize, False))

    fv = objective(pt)
    minima_list.append((pt, fv, True))

    def minima_callback(x, fv, accept):
        # print(x, fv, accept)
        minima_list.append((x, fv, accept))

    # perform the basin hopping search
    return basinhopping(objective, pt, take_step=take_step, niter=18, callback=minima_callback)

history_pos = 0
for k in range(len(init_points)):
    result = solve(k)
    # modify the first step
    first_step = takestep_history[history_pos]
    first_step.ret_x = takestep_history[history_pos+1].x
    print("first_step.x=", first_step.x, "first_step.ret_x=", first_step.ret_x)

    # add None to make the history size equal to that of minima_list
    takestep_history.append(None)

    history_pos = len(takestep_history)
    # summarize the result
    print('Status : %s' % result['message'])
    print('Total Evaluations: %d' % result['nfev'])
    # evaluate solution
    solution = result['x']
    evaluation = objective(solution)
    print('Solution: f(%s) = %.5f' % (solution, evaluation))

# sample input range uniformly at 0.1 increments
xaxis = arange(r_min, r_max, 0.1)
yaxis = arange(r_min, r_max, 0.1)
# create a mesh from the axis
xx, yy = meshgrid(xaxis, yaxis)
# compute targets
results = objective((xx, yy)) 
# create a surface plot with the jet color scheme
fig = plt.figure(figsize=(16,8))
ax = fig.add_subplot(121, projection='3d')
ax2 = fig.add_subplot(122)

fig.suptitle("Demonstration of Basin Hopping Algorithm using Ackley function", fontsize=20)
ax.set_title("Surface Plot", fontsize=16)
ax2.set_title("Contour Plot", fontsize=16)
fig.tight_layout()

surface = ax.plot_surface(xx, yy, results, cmap='jet', alpha=0.3)
contuor = ax2.contourf(xx, yy, results, cmap='jet', alpha=0.3)
# show the plot

minima_array = np.array([np.concatenate([x, [v]]) for x, v, a in minima_list])
x, y, z = minima_array[0]
points, = ax.plot([x], [y], [z], 'o', color='cyan')
trace3, = ax.plot(*minima_array[0:1,:].T, color='red')
point, = ax2.plot(x, y, 'o', color='cyan')
trace2, = ax2.plot(*minima_array[0:1,0:2].T, color='red', label="local minima trace")
first_step = takestep_history[0]
patch = first_step.create_patch()
ax2.add_patch(patch)

step_points = np.array([(np.nan, np.nan) if h is None else h.x for h in takestep_history])
DRAW_STEP_TRACE = False

if DRAW_STEP_TRACE:
    step_trace, = ax2.plot(*step_points[0:1].T, "-o", color="green", alpha=0.5)
displace = first_step.create_displine(ax2)

min_init, = ax2.plot(*step_points[0], "o", color="cyan", label="last initialization")
min_vector, = ax2.plot(*step_points[0:1].T, color="cyan", alpha=0.5, label="last minimization")
minimizatin = min_init, min_vector

ax2.legend()

print("len(takestep_history)=", len(takestep_history))
print("len(minima_list)=", len(minima_list))

def reset():
    x, y, z = minima_array[0]
    points.set_data(np.array([x]), np.array([y]))
    points.set_3d_properties(np.array([z]), 'z')
    trace3.set_data(*minima_array[0:1,0:2].T)
    trace3.set_3d_properties(minima_array[0:1,2], 'z')
    point.set_data([x], [y])
    trace2.set_data(*minima_array[0:1, 0:2].T)
    first_step = takestep_history[0]
    first_step.move_patch(patch)
    first_step.move_displine(displace)
    min_init.set_data([x], [y])
    min_vector.set_data(*step_points[0:1].T)
    if DRAW_STEP_TRACE:
        step_trace.set_data(*step_points[0:1].T)
        return points, trace3, point, trace2, patch, step_trace, *displace, *minimizatin
    else:
        return points, trace3, point, trace2, patch, *displace, *minimizatin

def animate(i):
    global step_arrow

    x, y, z = minima_array[i]
    points.set_data(np.array([x]), np.array([y]))
    points.set_3d_properties(np.array([z]), 'z')
    point.set_data([x], [y])
    color = 'cyan' if i % 20 == 0 else 'red'
    points.set_color(color)
    point.set_color(color)
    j, r = divmod(i, 20)
    start = j*20
    stop = start + r + 1
    slice_ = slice(start, stop)
    trace3.set_data(*minima_array[slice_,0:2].T)
    trace3.set_3d_properties(minima_array[slice_,2], 'z')
    trace2.set_data(*minima_array[slice_,0:2].T)
    this_step = takestep_history[i]
    if this_step is not None:
        this_step.move_patch(patch)
        this_step.move_displine(displace)
        if r > 0:
            x_, y_ = takestep_history[i-1].ret_x
            min_init.set_data([x_], [y_])
            min_vector.set_data([x_, x], [y_, y])
    if DRAW_STEP_TRACE:
        step_trace.set_data(*step_points[slice_].T)
        return points, trace3, point, trace2, patch, step_trace, *displace, *minimizatin
    else:
        return points, trace3, point, trace2, patch, *displace, *minimizatin

anim = animation.FuncAnimation(fig, animate, frames=len(minima_list), blit=True, init_func=reset, interval=500)
import sys
if len(sys.argv) > 1 and sys.argv[1] == "save":
    anim.save("improved-anim.gif")

plt.show()



# coding: utf-8
"""
    PdbFetcherDialog.py

    Copyright (c) 2020-2021, SAXS Team, KEK-PF
"""
import os
import sys
import re
import logging
import queue
from molass_legacy.KekLib.KillableThread import Thread
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ScrolledText
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.BasicUtils import is_empty_dir, mkdirs_with_retry
from molass_legacy.DENSS.DenssUtils import run_pdb2mrc

count_re = re.compile(r'(\d+)$')

def get_new_folder(folder):
    while os.path.exists(folder) and is_empty_dir(folder):
        folder = re.sub(count_re,  lambda m: str(int(m.group(1))+1), folder)
    return folder

"""
    Unifying these logging procesures with those in DenssGui
    seems to be a challenging task,
    because simply moving them to a module file beaks the logging
    from the separate thread.
    See also DenssFitData.py
    for an example of logging in the same thread.
"""
class StdoutRedirector:
    def __init__(self, queue):
        self.saved_stdout = sys.stdout
        self.queue = queue
        sys.stdout = self

    def write(self, string):
        self.queue.put([1, string])

    def flush(self):
        pass

    def __del__(self):
        sys.stdout = self.saved_stdout
        print('__del__ ok')

class ProgressLoghandler(logging.Handler):
    def __init__(self, log_text, log_file):
        logging.Handler.__init__(self)
        self.log_text   = log_text
        self.fh = open(log_file, 'w')
        self.formatter  = logging.Formatter('%(asctime)s %(message)s', '%H:%M:%S')

    def emit( self, record ):
        if record.levelno >= logging.INFO:
            try:
                log_message = self.formatter.format(record) + '\n'
                self.fh.write(log_message)
                self.fh.flush()
                self.log_text.insert(Tk.INSERT, log_message)
                self.log_text.see(Tk.END)
            except:
                # this error can occur when canceled
                pass

    def __del__(self):
        self.fh.close()

class PdbFetcherDialog(Dialog):
    def __init__(self, parent):
        self.applied = False
        self.mrc_file = None
        Dialog.__init__( self, parent, "PDB Fetcher", visible=False )

    def show(self):
        self._show()

    def body(self, body_frame):
        input_frame = Tk.Frame(body_frame)
        input_frame.pack(padx=20, pady=10)
        log_frame = Tk.Frame(body_frame)
        log_frame.pack(padx=20, pady=10)

        entry_width = 50
        grid_row = 0
        label = Tk.Label(input_frame, text="PDB ID: ")
        label.grid(row=grid_row, column=0, sticky=Tk.E)
        self.search_key = Tk.StringVar()
        entry = Tk.Entry(input_frame, textvariable=self.search_key, width=entry_width)
        entry.grid(row=grid_row, column=1, sticky=Tk.W)
        w = Tk.Button(input_frame, text="Fetch", width=10, command=self.start_fetch_thread)
        w.grid(row=grid_row, column=2, padx=20, pady=5)
        self.fetch_btn = w

        grid_row += 1
        label = Tk.Label(input_frame, text="Out Folder: ")
        label.grid(row=grid_row, column=0, sticky=Tk.E)
        self.out_folder = Tk.StringVar()
        out_folder = get_setting("analysis_folder") + r"\PDB\000"
        out_folder = get_new_folder(out_folder)
        self.out_folder.set(out_folder)
        entry = FolderEntry( input_frame, textvariable=self.out_folder, width=entry_width,
                            slimbutton=True)
        entry.grid(row=grid_row, column=1, sticky=Tk.W)

        self.log_text = ScrolledText(log_frame, width=78, height=10 )
        self.log_text.pack()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w.pack_forget()
        self.ok_btn = w

        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.cancel_btn = w

        self.bind("<Escape>", self.cancel)

    def apply(self):
        print('apply')
        self.applied = True

    def check_module(self):
        try:
            import requests
            import pypdb
            import mrcfile
            ok = True
        except:
            ok = False

        if not ok:
            from UpdateDailog import UpdateDailog
            modules = ['requests', 'pypdb', 'mrcfile']
            dialog = UpdateDailog(self, modules)
            dialog.show()
            if dialog.applied:
                ok = True

        return ok

    def start_fetch_thread(self):
        ok = self.check_module()
        if not ok:
            return

        from molass_legacy.KekLib.BasicUtils import get_home_folder
        self.fetch_btn.config(state=Tk.DISABLED)

        out_folder = self.out_folder.get()
        out_folder = get_new_folder(out_folder)
        self.out_folder.set(out_folder)

        log_folder = get_home_folder() + '/log'
        log_file = log_folder + '/fetcher.log'
        self.root_logger  = logging.getLogger()
        self.root_logger.addHandler( ProgressLoghandler(self.log_text, log_file) )
        self.queue = queue.Queue()
        self.thread = Thread(
                        target=self.fetch,
                        name='DenssThread',
                        args=[]
                        )
        self.thread.start()
        self.counter = 0
        self.maximum = None
        self.last_index = self.log_text.index(Tk.INSERT)
        self.after(200, self.update_progress_text)      # 200: need some time to have self.log_text written enough

    def update_progress_text(self):
        try:
            while True:
                ret = self.queue.get(block=False)
                if ret[0] == 0:
                    self.thread.join()
                    self.cancel_btn.pack_forget()
                    self.ok_btn.pack(side=Tk.LEFT, padx=5, pady=5)
                    self.fetch_btn.config(state=Tk.NORMAL)
                    return
                else:
                    message = ret[1]
                    if self.counter > 0:
                        self.log_text.delete( self.last_index, Tk.INSERT )

                    self.last_index = self.log_text.index(Tk.INSERT)
                    if len(message) > 0 and message[0] == '\r':
                        self.counter += 1
                    else:
                        # self.counter = 0
                        pass
                    self.log_text.insert(Tk.INSERT, message)

        except queue.Empty:
            pass
        self.after(100, self.update_progress_text)

    def fetch(self):
        from time import sleep
        from .PypdbLite import Query, get_pdb_file
        self.redirector = StdoutRedirector(self.queue)

        pdbid = self.search_key.get()
        print("Query('%s')" % pdbid)
        results = Query(pdbid).search()
        self.queue.put((1, ""))

        if len(results) > 0:
            out_folder = self.out_folder.get()
            mkdirs_with_retry(out_folder)
            ret = get_pdb_file(pdbid, filetype='pdb')
            # print(ret)
            pdb_file = os.path.join(out_folder, pdbid + '.pdb')
            try:
                with open(pdb_file, "w") as fh:
                    fh.write(ret)
            except Exception as exc:
                print("Error", exc)
                return
            print("Saved to %s" % pdb_file)
            self.queue.put((1, ""))
            self.mrc_file = run_pdb2mrc(pdb_file, queue=self.queue)
            if self.mrc_file is not None:
                print("It is ready. Press OK.")
        else:
            print("Not found.")

        self.queue.put((0, ""))

    def get_mrc_file_path(self):
        return self.mrc_file



"""
    PathUtils.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import os
import re

class NonExistingPath:
    def __init__(self, path):
        while os.path.exists(path):
            path = re.sub(r"^([^\(\)]+\S)(\s+)?(\(\d+\))?$", self.new_name, path)

        self._path = path

    def new_name(self, m):
        m3 = m.group(3)
        if m3 is None:
            m3 = "0"
        else:
            m3 = m3[1:-1]
        return m.group(1) + " " + "(%d)" % (int(m3)+1)

    def __str__(self):
        return self._path



# coding: utf-8
"""
    AutoRangeInspector.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import sys
from time import sleep
import queue
import logging
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.ReadOnlyText import ReadOnlyText
from molass_legacy.KekLib.KillableThread import Thread
from TkTester import AppAgent

class StdoutRedirector:
    def __init__(self, queue):
        self.saved_stdout = sys.stdout
        self.queue = queue
        sys.stdout = self

    def write(self, string):
        self.queue.put([1, string])

    def flush(self):
        pass

    def __del__(self):
        sys.stdout = self.saved_stdout
        print('__del__ ok')

class AutoRangeInspector(Dialog):
    def __init__(self, parent, inspector):
        self.inspector = inspector
        Dialog.__init__(self, parent, "Auto Range Inspector", visible=False)

    def body(self, body_frame):
        self.log_text = ReadOnlyText(body_frame, width=80, height=10)
        self.log_text.pack()
        self.log_text.update()

    def inspect(self):
        self.after(100, self.inspect_start)
        self._show()

    def inspect_start(self):
        self.agent = AppAgent(self.inspector)
        self.queue = queue.Queue()
        self.thread = Thread(
                        target=self.inspect_thread,
                        name='InspectThread',
                        args=[]
                        )
        self.thread.start()

        self.watch()

    def watch(self):
        try:
            while True:
                ret = self.queue.get(block=False)
                if ret[0] == 0:
                    self.thread.join()
                    self.ok()
                    return
                else:
                    message = ret[1]
                    self.log_text.insert(Tk.INSERT, message)
                    self.log_text.see(Tk.END)
        except queue.Empty:
            pass

        self.after(100, self.watch)

    def inspect_thread(self):
        redirector = StdoutRedirector(self.queue)

        for k in range(10):
            print([k])
            self.agent.debug_print(str([k]))
            sleep(1)

        logging.info('Inspect thread finished.')
        self.queue.put([0])



# coding: utf-8
"""

    MessageBoxUtils.py

    Copyright (c) 2019, Masatsuyo Takahashi, KEK-PF

"""

"""
    References:
        (1) Win32 Python: Getting all window titles
            https://sjohannes.wordpress.com/tag/win32/

        (2) MessageBoxをプログラムから閉じたい
            http://tarulab.blogspot.com/2014/03/messagebox.html

        (3) 第２３章 ダイアログボックスの研究 その１
            http://mrxray.on.coocan.jp/Halbow/Chap23.html
"""
import ctypes

EnumWindows = ctypes.windll.user32.EnumWindows
EnumWindowsProc = ctypes.WINFUNCTYPE(ctypes.c_bool, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))
GetWindowText = ctypes.windll.user32.GetWindowTextW
GetWindowTextLength = ctypes.windll.user32.GetWindowTextLengthW
IsWindowVisible = ctypes.windll.user32.IsWindowVisible

WM_COMMAND = 0x0111
ID_OK = 1
ID_CANCEL = 2
ID_ABORT = 3
ID_RETRY = 4
ID_IGNORE = 5
ID_YES = 6
ID_NO = 7
ID_CLOSE = 8
ID_HELP = 9

COMMAND_DICT = {'Y':ID_YES, 'N':ID_NO, 'O':ID_OK, 'C':ID_CANCEL}

class Window:
    def __init__(self, hwnd, title):
        self.hwnd = hwnd
        self.title = title

    def send_command(self, command_id):
        ctypes.windll.user32.SendMessageW(self.hwnd, WM_COMMAND, command_id, 0)

def find_windows(pattern):
    windows = []   
    def foreach_window(hwnd, lParam):
        if IsWindowVisible(hwnd):
            length = GetWindowTextLength(hwnd)
            buff = ctypes.create_unicode_buffer(length + 1)
            GetWindowText(hwnd, buff, length + 1)
            if buff.value.find(pattern) >=0 :
                windows.append(Window(hwnd, buff.value))
        return True

    EnumWindows(EnumWindowsProc(foreach_window), 0)

    return windows

def reply_messagebox(title, reply):
    windows = find_windows(title)
    assert len(windows) == 1
    command_id = COMMAND_DICT[reply]
    windows[0].send_command(command_id)

def window_exists(pattern):
    return len(find_windows(pattern)) > 0

if __name__ == '__main__':
    # show a messagebox titled "Test Title"
    # and try one of the following
    reply_messagebox('Test Title', 'Y')     # when showing Info
    reply_messagebox('Test Title', 'Y')     # when asking Yes/No
    reply_messagebox('Test Title', 'O')     # when asking Ok/Cancel



"""
    Optimizer.FunctionDebuggerUtils.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
from molass_legacy.Baseline.BaselineUtils import create_xr_baseline_object
from .FuncImporter import import_objective_function

def create_optimizer_for_debug(dsets, n_components, optinit_info, init_params, composite=None, prepare=True):
    print("create_optimizer_for_debug: prepare=", prepare)

    # task-remark-begin unify the coding below with Optimizer.JobStateInfo.py
    uv_base_curve = optinit_info.treat.get_base_curve()
    class_code = optinit_info.class_code
    fullopt_class = import_objective_function(class_code)
    sd = optinit_info.sd
    xr_base_curve = create_xr_baseline_object()
    new_optimizer = fullopt_class(dsets, n_components,
                                uv_base_curve=uv_base_curve,
                                xr_base_curve=xr_base_curve,
                                qvector=sd.qvector,
                                wvector=sd.lvector,
                                composite=composite,
                                )
    if prepare:
        new_optimizer.prepare_for_optimization(init_params)
    # task-remark-end

    return new_optimizer



"""
    RangeInspector.py

    Copyright (c) 2020-2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Rectangle
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from molass_legacy.GuinierAnalyzer.AutorgKekAdapter import AutorgKekAdapter
from molass_legacy.Env.EnvInfo import get_global_env_info
from molass_legacy.KekLib.SaferSpinbox import SaferSpinbox 
from molass_legacy.KekLib.ExceptionTracebacker import log_exception

TITLE_FONTSIZE = 16
AVERAGE_WIDTH = 10
NSRD_INDEX = 2
SCALE_TYPE_PARAMS = [("log", 0), ("linear", 1)]
LINE_TYPE_NAMES = ["peaktop", "init", "kept", "almerge"]
LINE_TYPE_IVARS = [1, 1, 1, 0]
ALMERGE_ROW = -1

class RangeInspectorDialog(Dialog):
    def __init__(self, parent, solver, j0, paired_range, selected, ad, conc_depend=None, init_result=None, decomp_info=None, fx=None):
        assert conc_depend is not None

        self.solver = solver
        self.q = solver.qv
        self.zero_diff = np.zeros(len(self.q))
        self.data = solver.data
        self.error = solver.error
        self.x_curve = solver.ecurve
        self.mc_vector = solver.mc_vector

        enf_info = get_global_env_info()
        self.atsas_ok = enf_info.atsas_is_available
        self.almerge_executer = None
        self.almerge_y = None

        self.show_diff_mode = Tk.IntVar()
        self.parent = parent
        self.j0 = j0
        self.paired_range = paired_range
        self.selected = selected    # selected peakset_info
        self.ad = ad
        self.conc_depend = conc_depend
        self.scale_type = Tk.IntVar()
        self.compute_peak_top_scattering()
        self.create_line_variables()

        f, t = self.get_initial_range()
        if init_result is None:
            A, E = self.solve_this_range(f, t+1)
            # self.C has been set in the above call
        else:
            A, B, Z, E_, _, C = init_result
            E = E_[0]
            self.C = C
            self.compute_almerge_scattering(f, t+1)
        self.A_init = A
        self.E_init = E
        self.nSRD_init = self.compute_nSRD_info(A, E, f, t+1)
        self.max_nsrd = self.nSRD_init[NSRD_INDEX]*10
        self.Rg_init = self.compute_rg_info(A, E)
        if decomp_info is None:
            self.scaled_recs = None
            self.fx = None
        else:
            conc_factor = solver.conc_factor
            self.scaled_recs = decomp_info.get_scaled_recs(conc_factor)
            self.fx = decomp_info.fx
        self.create_range_variables()
        self.low_quality_warned = False
        self.applied = False
        Dialog.__init__( self, parent, "Range Inspector", visible=False)

    def compute_peak_top_scattering(self):
        top_ = self.paired_range.top_x
        hw = AVERAGE_WIDTH//2
        y = np.average(self.data[:,top_-hw:top_+hw], axis=1)
        self.top_c = self.mc_vector[top_]
        self.pty = y_  = y / self.top_c
        error = np.average(self.error[:,top_-hw:top_+hw], axis=1)
        self.pt_sn = self.pty/error
        Y = y_
        wider_range_start = int(len(y)*0.7)
        Yw = Y[wider_range_start:]
        Yw_ = Yw[np.isfinite(Yw)]
        self.min_log_y = np.average(Yw_) - np.std(Yw_)
        self.peak_top_x = top_ + self.j0
        self.peak_top_y = Y
        self.j_min = self.j0
        self.j_max = self.j0 + len(y) - 1

    def get_initial_range(self):
        fromto_list = self.paired_range.get_fromto_list()
        fv, tv = [ self.j0 + j  for j in fromto_list[self.ad]]
        return fv, tv

    def create_range_variables(self):
        fv, tv = self.get_initial_range()
        print('paired_range=', self.paired_range, (fv, tv))

        self.init_interval = (fv, tv)
        self.show_interval = [fv, tv]   # using a list to allow asignment
        f = Tk.IntVar()
        f.set(fv)
        t = Tk.IntVar()
        t.set(tv)
        self.range_vars = [f, t]

    def create_line_variables(self):
        self.line_vars = []
        for val in LINE_TYPE_IVARS:
            ivar = Tk.IntVar()
            ivar.set(val)
            self.line_vars.append(ivar)

    def show(self):
        self._show()
        return self.applied

    def body(self, body_frame):
        canvas_frame = Tk.Frame(body_frame)
        canvas_frame.pack()
        panel_frame = Tk.Frame(body_frame)
        panel_frame.pack(fill=Tk.X)
        tframe = Tk.Frame(panel_frame)
        tframe.pack(side=Tk.LEFT)
        dframe = Tk.Frame(panel_frame)
        dframe.pack(side=Tk.RIGHT)
        pframe = Tk.Frame(panel_frame)
        pframe.pack(side=Tk.RIGHT)

        self.build_canvas(canvas_frame, tframe)
        self.build_panel(pframe)
        self.builf_distinction(dframe)

    def build_canvas(self, canvas_frame, tool_frame):
        figsize = (23,7)
        fig = plt.figure(figsize=figsize)
        nrows, ncols = 1, 3
        height_ratios = [5, 1]
        gs = GridSpec( 2, ncols, height_ratios=height_ratios )
        self.height_ratio = height_ratios[1]/height_ratios[0]

        ax1 = fig.add_subplot(gs[:,0])
        ax2 = fig.add_subplot(gs[0,1])
        ax2.get_xaxis().set_visible(False)
        ax2d = fig.add_subplot(gs[1,1])
        ax3 = fig.add_subplot(gs[:,2])
        self.ax2d = ax2d
        self.ax3t = ax3.twinx()
        self.axes = [ax1, ax2, ax2d, ax3, self.ax3t]

        self.mpl_canvas = FigureCanvasTkAgg(fig, canvas_frame)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tool_frame)
        self.toolbar.update()
        # self.mpl_canvas.mpl_connect('motion_notify_event', self.on_mouse_motion)

        in_folder = get_in_folder()
        fig.suptitle("Range Dependency Inspection for " + in_folder, fontsize=20)

        axt = ax1.twinx()
        axt.grid(False)
        self.draw_elution(ax1, axt)
        self.draw_scattering(ax2)

        fig.tight_layout()
        fig.subplots_adjust(top=0.88, wspace=0.3, left=0.05, right=0.95)

        self.mpl_canvas.draw()
        self.fig = fig
        self.axes = [ax1, ax2, ax3]
        self.place_scale_buttons(canvas_frame)
        self.place_line_buttons(canvas_frame)
        self.place_diff_toggle_button(canvas_frame)

    def place_scale_buttons(self, frame):
        button_frame = Tk.Frame(frame)
        button_frame.place(relx=0.63, rely=0.13)
        for name, value in SCALE_TYPE_PARAMS:
            rb = Tk.Radiobutton(button_frame, text=name, variable=self.scale_type, value=value)
            rb.pack(anchor=Tk.W)
        self.scale_type.trace("w", self.scale_type_tracer)

    def scale_type_tracer(self, *args):
        self.update_yscale(self.axes[1])
        self.mpl_canvas.draw()

    def update_yscale(self, ax):
        ax.set_yscale("linear" if self.scale_type.get() else "log")

    def place_line_buttons(self, frame):
        button_frame = Tk.Frame(frame)
        button_frame.place(relx=0.63, rely=0.3)
        for name, ivar in zip(LINE_TYPE_NAMES, self.line_vars):
            if name == "almerge":
                state = Tk.NORMAL if self.atsas_ok else Tk.DISABLED
            else:
                state = Tk.NORMAL
            cb = Tk.Checkbutton(button_frame, text=name, variable=ivar, state=state)
            cb.pack(anchor=Tk.W)
            ivar.trace("w", self.line_buttons_tracer)

    def line_buttons_tracer(self, *args):
        if self.line_vars[ALMERGE_ROW].get() == 1:
            self.update_almerge_line()

        for ivar, lines in zip(self.line_vars, self.line_table):
            for line in lines:
                if line is not None:
                    line.set_visible(ivar.get() == 1)
        self.mpl_canvas.draw()

    def place_diff_toggle_button(self, frame):
        switch_frame = Tk.Frame(frame)
        switch_frame.place(relx=0.63, rely=0.8)
        for k, t in enumerate(["PT-diff", "SR-diff"]):
            rb = Tk.Radiobutton(switch_frame, text=t, variable=self.show_diff_mode, value=k)
            rb.pack()
        self.show_diff_mode.trace("w", self.show_diff_mode_tracer)

    def show_diff_mode_tracer(self, *args):
        self.range_ends_tracer(self.ad)

    def draw_elution(self, ax, axt):
        ax.set_title("Elution Curves", fontsize=TITLE_FONTSIZE)
        ax.set_xlabel("Elution №")
        ax.set_ylabel("Concentration (Absorbance x Conc. Factor)")
        axt.set_ylabel("Scattering Intensity")
        curve = self.x_curve
        x = self.j0 + curve.x
        fx = self.fx
        y = curve.y
        ax.plot(x, self.mc_vector, color='blue', alpha=0.5, label='measured UV')
        axt.plot(x, y, color='orange', alpha=0.5, label='measured Xray')

        if self.scaled_recs is not None:
            for k, rec in enumerate(self.scaled_recs):
                func = rec[1]
                y_ = func(fx)
                ax.plot(x, y_, ':', label='Scaled Xray component-%d' % k)

        ax.legend(bbox_to_anchor=(0, 1.0), loc='upper left')
        axt.legend(loc='upper right')

        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)

        fv, tv = [var.get() for var in self.range_vars]
        self.rect_patch = Rectangle(
                (fv, ymin),  # (x,y)
                tv - fv,   # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax.add_patch(self.rect_patch)

        lines = []
        for x in [fv, tv]:
            line, = ax.plot( [x, x], [ymin, ymax], ':', color='gray', alpha=0.5 )
            lines.append(line)
        self.lines = lines
        self.elution_ylim = (ymin, ymax)
        self.last_range = (fv, tv)

    def draw_scattering(self, ax):
        self.init_control_vars()

        ax.cla()
        self.update_yscale(ax)
        ax.set_title("Scattering Curves", fontsize=TITLE_FONTSIZE)
        ax.set_ylabel("Intensity")
        self.scatter_x = self.q

        line, = ax.plot(self.scatter_x, self.peak_top_y, alpha=0.5, label='raw data average near peak top at %d' % self.peak_top_x)
        self.line_table.append([line])

        fv, tv = self.get_initial_range()
        line, = ax.plot(self.scatter_x, self.A_init, alpha=0.5, label='extrapolated with initial range (%d, %d)' % (fv, tv))
        self.line_table.append([line])

        self.line_table.append(self.scattering_curves)  # for kept lines
        self.line_table.append([None])  # for almerge

        _, ymax = ax.get_ylim()
        # ax.set_ylim(self.min_log_y, ymax)
        ax.legend()
        ax2d = self.ax2d
        ax2d.cla()
        ax2d.set_ylabel("diff")
        ax2d.set_xlabel("Q")
        m = self.show_diff_mode.get()
        self.diff_line, = ax2d.plot(self.q, self.nSRD_init[m], color='C1')

    def update_almerge_line(self):
        if not self.atsas_ok:
            return
        if not self.line_vars[ALMERGE_ROW].get():
            return

        if self.almerge_y is None:
            f, t = [var.get() - self.j0 for var in self.range_vars]
            self.compute_almerge_scattering(f, t+1)

        x = self.scatter_x
        y = self.almerge_y
        line = self.line_table[ALMERGE_ROW][0]
        if line is None:
            ax = self.axes[1]
            line, = ax.plot(x, y, ':', color='red', label='almerge')
            self.line_table[ALMERGE_ROW][0] = line
        else:
            line.set_data(x, y)

    def init_control_vars(self):
        self.line_table = []
        self.scattering_curve = None
        self.scattering_curves = []
        self.manip_infos = []
        self.manip_history = []
        self.last_manip = (self.Rg_init[0], self.A_init, *self.init_interval, *self.Rg_init[3:6], True)
        self.show_interval = list(self.init_interval)

    def build_panel(self, panel_frame):
        range_frame = Tk.Frame(panel_frame)
        range_frame.pack(side=Tk.LEFT, padx=10)

        spinbox_label = Tk.Label(range_frame, text="Range: ")
        spinbox_label.pack(side=Tk.LEFT)

        spinbox_frame = Tk.Frame(range_frame)
        spinbox_frame.pack(side=Tk.LEFT)

        f, t = self.range_vars
        spinbox1 = SaferSpinbox( spinbox_frame, textvariable=f,
                            from_=self.j_min, to=self.j_max, increment=1,
                            justify=Tk.CENTER, width=6 )
        spinbox1.pack(side=Tk.LEFT, padx=5, pady=5)
        spinbox1.set_tracer(lambda *args: self.range_ends_tracer(0))
        f.trace('w', spinbox1.tracer)

        spinbox2 = SaferSpinbox( spinbox_frame, textvariable=t,
                            from_=self.j_min, to=self.j_max, increment=1,
                            justify=Tk.CENTER, width=6 )
        spinbox2.pack(side=Tk.LEFT, padx=5, pady=5)
        spinbox2.set_tracer(lambda *args: self.range_ends_tracer(1))
        t.trace('w', spinbox2.tracer)

        self.range_ends_tracing  = True

        length_frame = Tk.Frame(range_frame)
        length_frame.pack(side=Tk.LEFT, padx=5)

        self.fixed_length = Tk.IntVar()
        self.fixed_length.set(0)
        self.fixed_length_cb = Tk.Checkbutton(length_frame, variable=self.fixed_length, text="Fixed Length:")
        self.fixed_length_cb.pack(side=Tk.LEFT)

        fv, tv = [w.get() for w in self.range_vars]
        self.range_length = Tk.IntVar()
        self.range_length.set(tv - fv + 1)
        length_label = Tk.Label(length_frame, textvariable=self.range_length, width=3)
        length_label.pack(side=Tk.LEFT)

        self.keep_btn = Tk.Button(range_frame, text="Keep", command=self.keep_scattering_curve, state=Tk.DISABLED)
        self.keep_btn.pack(side=Tk.LEFT, padx=5)

        self.clear_btn = Tk.Button(range_frame, text="Clear", command=self.clear_scattering_curves, state=Tk.DISABLED)
        self.clear_btn.pack(side=Tk.LEFT, padx=5)

    def builf_distinction(self, dframe):
        self.plot_distinction(legend=True)
        self.update()
        canvas_width = int(self.mpl_canvas_widget.cget( 'width' ))//3
        space_width = canvas_width//2
        space = Tk.Frame(dframe, width=space_width)
        space.pack(side=Tk.LEFT)
        auto_btn = Tk.Button(dframe, text="Guinier/Kratky Plots", command=self.show_guinier_kratky_plots)
        auto_btn.pack(side=Tk.LEFT)
        space = Tk.Frame(dframe, width=space_width)
        space.pack(side=Tk.LEFT)

    def plot_distinction(self, legend=False):
        ax = self.axes[2]
        ax.set_title("Supporting Information (nSRD, Rg)", fontsize=TITLE_FONTSIZE)
        ax.set_ylabel("nSRD")
        ax.set_xlabel("Elution №")
        x = self.init_interval[self.ad]
        ax.plot(x, self.nSRD_init[NSRD_INDEX], 'o', color='C1')
        ax3t = self.ax3t
        ax3t.set_ylabel("Rg")
        ax3t.grid(False)
        self.plot_rg_errorbar(ax3t, x, self.Rg_init, color='C1', label='initial Rg (%d,%d)' % self.init_interval)
        if legend:
            self.layout_distinction()

    def plot_rg_errorbar(self, ax, x, Rg_rec, color=None, label=None):
        rg, rg_error, basic_quality = Rg_rec[0:3]
        if basic_quality < 0.5:
            if not self.low_quality_warned:
                self.after(2000, self.show_quality_warning)
            ax.plot(x, rg, 'o', color='red', alpha=0.5, markersize=20)
        ax.errorbar(x, rg, rg_error, fmt='o', markersize=5, capsize=3, elinewidth=3, mfc='black', color=color, label=label)

    def show_quality_warning(self):
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        self.update()
        if self.conc_depend == 1:
            rank_mismatch_message = ".\n"
        else:
            rank_mismatch_message = (
                ', and that "low quality"\nmay have been caused by rank mismatch,\n'
                "in that case, consider retrying with Conc. Dependency=1."
                )
        MessageBox.showwarning("Guinier Analysis Quality Warning",
            "Be aware that low quality Rg values are marked\n"
            "with a red circle"
            + rank_mismatch_message,
            parent=self,
            )
        self.low_quality_warned = True

    def layout_distinction(self):
        ax3 = self.axes[2]
        ax3.set_ylim(-self.max_nsrd*0.05, self.max_nsrd)
        loc = 'lower right' if self.ad == 0 else 'lower left'
        ax3.legend(loc=loc)

        ax3t = self.ax3t
        ax3t.set_xlim(self.show_interval[0]-2, self.show_interval[1]+2)
        ax3t.legend()

    def range_ends_tracer(self, i):
        self.update()
        if not self.range_ends_tracing:
            return

        fixed_length = self.fixed_length.get()
        # print([i], 'fixed_length=', fixed_length)
        error = False
        try:
            self.range_ends_tracing = False
            fv, tv = [var.get() for var in self.range_vars]
            if fv < tv:
                if fixed_length:
                    range_length = self.range_length.get()
                    i_ = 1 - i
                    if i_ == 0:
                        self.range_vars[i_].set(tv + 1 - range_length)
                    else:
                        self.range_vars[i_].set(fv + range_length - 1)
                    self.update()
                else:
                    self.range_length.set(tv - fv + 1)
            else:
                for var, val in zip(self.range_vars, self.last_range):
                    var.set(val)
        except:
            # as in cases where the values are cleared
            error = True

        self.update()

        if error:
            self.range_ends_tracing = True
            return

        try:
            self.update_drawing(fv, tv)

            if len(self.scattering_curves) == 1:
                for w in [self.keep_btn, self.clear_btn]:
                    w.config(state=Tk.NORMAL)

            self.last_range = (fv, tv)
            self.range_ends_tracing = True
        except:
            log_exception(None, "error in range_ends_tracer:")

    def update_drawing(self, fv, tv):
        patch = self.rect_patch
        x, y = patch.xy
        patch.set_xy((fv, y))
        patch.set_width(tv - fv)

        ax1, ax2, ax3 = self.axes
        ymin, ymax = ax1.get_ylim()

        for line, x_ in zip(self.lines, [fv, tv]):
            line.set_data([x_, x_], [ymin, ymax])

        A, E = self.solve_this_range(fv, tv+1)
        scatter_y = A

        label='extrapolated with new range (%d, %d)' % (fv, tv)
        if self.scattering_curve is None:
            line, = ax2.plot(self.scatter_x, scatter_y, label=label)
            self.scattering_curve = line
            self.scattering_curves.append(line)
            self.manip_infos.append(None)
        else:
            self.scattering_curve.set_data(self.scatter_x, scatter_y)
            self.scattering_curve.set_label(label)

        self.update_current_manip_info(A, E)

        ax2.legend()
        self.mpl_canvas.draw()

    def solve_this_range(self, f, t, surplus_rank=0):
        start, stop = [j - self.j0 for j in [f, t]]
        A, B, Z, E, _, C = self.solver.extrapolate_wiser(start, stop, self.selected,
                                conc_depend = self.conc_depend,
                                surplus_rank=surplus_rank)
        self.C = C
        self.compute_almerge_scattering(start, stop)
        return A, E[0]

    def compute_nSRD_info(self, A, E, f, t):
        AE = A/E
        A_, E_ = self.solve_this_range(f, t, surplus_rank=1)
        sr_diff = A_/E_ - AE
        pt_diff = AE - self.pt_sn
        return pt_diff, sr_diff, np.sqrt(np.average(sr_diff**2))

    def update_diff_info(self, diff):
        self.diff_line.set_data(self.q, diff)
        k = len(self.scattering_curves)
        self.diff_line.set_color('C%d' % (k+1))
        self.ax2d.set_ylim(diff.min(), diff.max())

    def compute_rg_info(self, A, E):
        A_data = np.vstack( [self.q, A, E] ).T
        autorg_kek = AutorgKekAdapter( A_data )
        result = autorg_kek.run()
        if result.Rg is None:
            ret = (np.nan, np.nan, np.nan, np.nan, np.nan, np.nan)
        else:
            ret = (result.Rg, result.Rg_stdev, result.basic_quality, result.From, result.To, result.I0)
        return ret

    def update_current_manip_info(self, A=None, E=None):
        ax3 = self.axes[2]
        for ax in [ax3, self.ax3t]:
            ax.cla()

        self.plot_distinction()
        ax = self.ax3t

        if A is not None:
            fv, tv = [ self.range_vars[k].get() for k in range(2) ]
            self.show_interval[0] = min(fv, self.show_interval[0])
            self.show_interval[1] = max(tv, self.show_interval[1])
            Rg, Rg_stdev, basic_quality, gf, gt, I0 = self.compute_rg_info(A, E)
            nSRD = self.compute_nSRD_info(A, E, fv, tv+1)
            m = self.show_diff_mode.get()
            self.update_diff_info(nSRD[m])
            nsrd_ = nSRD[NSRD_INDEX]
            self.manip_infos[-1] = ((fv, tv), (Rg, Rg_stdev, basic_quality, nsrd_))
            if len(self.manip_history) == 0:
                self.manip_history.append((self.init_interval[self.ad], self.Rg_init[0], self.nSRD_init[NSRD_INDEX]))
            pos = fv if self.ad == 0 else tv
            self.manip_history.append((pos, Rg, nsrd_))
            self.last_manip = (Rg, A, fv, tv, gf, gt, I0, False)

            manip_array = np.array(self.manip_history)
            ax.plot(manip_array[:,0], manip_array[:,1], ':', color='red', alpha=0.5, label='Rg history')
            ax3.plot(manip_array[:,0], manip_array[:,2], ':', color='blue', alpha=0.5, label='nSRD history')

            for k, (interval, manip_info) in enumerate(self.manip_infos, start=2):
                x = interval[self.ad]
                color = 'C%d' % k
                ax3.plot(x, manip_info[3], 'o', color=color)
                self.plot_rg_errorbar(ax, x, manip_info[0:3], color=color, label='Rg (%d,%d)' % interval)

        self.layout_distinction()

    def keep_scattering_curve(self):
        self.scattering_curve = None

    def clear_scattering_curves(self):
        self.draw_scattering(self.axes[1])
        self.update_current_manip_info()
        self.mpl_canvas.draw()

    def validate(self):
        if self.scattering_curve is None:
            ret = True
        else:
            import molass_legacy.KekLib.CustomMessageBox as MessageBox
            interval, manip_info = self.manip_infos[-1]
            assert interval == self.last_range
            Rg, Rg_stdev, basic_quality = manip_info[0:3]
            if basic_quality < 0.5:
                low_quality_warning = "\nAnd Rg=%.1f has been estimated with a low basic_quality=%.3g.\n\n" % (Rg, basic_quality)
            else:
                low_quality_warning = ""
            ret = MessageBox.askyesno( "Change Confirmation",
                'Pressing "OK" here implies\n'
                + 'to apply the current changed range (%d, %d).\n' % self.last_range
                + low_quality_warning
                + "Are you sure to change it as stated?", parent=self.parent )
        return ret

    def apply(self):
        if self.scattering_curve is None:
            return

        print("apply")
        self.applied = True

    def get_range(self, shifted=True):
        if shifted:
            return self.last_range
        else:
            return [j - self.j0 for j in self.last_range]

    def auto_inspection(self):
        from .AutoRangeInspector import AutoRangeInspector
        ari = AutoRangeInspector(self, self)
        ari.inspect()

    def show_guinier_kratky_plots(self, debug=False):
        if debug:
            from importlib import reload
            import Kratky.GuinierKratkyPlots
            reload(Kratky.GuinierKratkyPlots)
        from Kratky.GuinierKratkyPlots import GuinierKratkyPlots
        rg, y, f, t, gf, gt, I0, initial = self.last_manip
        if rg is None:
            pass
        else:
            color = "C1" if initial else "C2"
            gp = GuinierKratkyPlots(self.parent, self.q, y, rg, I0, (f, t), (gf, gt), color=color)
            gp.show()

    def compute_almerge_scattering(self, start, stop):
        if not self.atsas_ok:
            return
        if not self.line_vars[ALMERGE_ROW].get():
            return

        if self.almerge_executer is None:
            from molass_legacy.ATSAS.Almerge import AlmergeExecutor
            self.almerge_executer = AlmergeExecutor()

        slice_ = slice(start, stop)
        M = self.data[:,slice_]
        E = self.error[:,slice_]
        c_vector = np.sum(self.C, axis=0)
        result = self.almerge_executer.execute_matrix(self.q, M, E, c_vector)
        print('result.exz_array.shape', result.exz_array.shape)
        self.almerge_y = result.exz_array[:,1]/self.top_c
        if False:
            import molass_legacy.KekLib.DebugPlot as dplt
            dplt.push()
            fig, ax = dplt.subplots()
            ax.set_yscale("log")
            ax.plot(self.q, self.almerge_y)
            fig.tight_layout()
            dplt.show()
            dplt.pop()

    def on_mouse_motion(self, event):
        # working on this
        print("on_mouse_motion", event.inaxes, self.axes[2], event.xdata, event.ydata)
        if event.inaxes == self.axes[2]:
            print("on_mouse_motion", event.xdata, event.ydata)



import numpy                as np
import matplotlib.pyplot    as plt

N   = 10
a   = 5
b   = 3
s   = 0.2
x   = np.linspace( 0, 1, N )
np.random.seed( 1234 )
e   = np.random.normal( 0, s, N )
y   = a * x + b + e

yt  = a * x + b

plt.plot( x, y, 'o' )
plt.plot( x, yt, ':r' )
plt.show()



from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import numpy as np
import matplotlib.pyplot as plt

point = np.array([1, 2, 3])
normal = np.array([1, 1, 2])

# a plane is a*x+b*y+c*z+d=0
# [a,b,c] is the normal. Thus, we have to calculate
# d and we're set
d = -point.dot(normal)

# create x,y
xx, yy = np.meshgrid(range(10), range(10))

# calculate corresponding z
z = (-normal[0] * xx - normal[1] * yy - d) * 1. / normal[2]

# plot the surface
plt3d = plt.figure().add_subplot(111, projection='3d')

Gx, Gy = np.gradient(xx * yy)  # gradients with respect to x and y
G = (Gx ** 2 + Gy ** 2) ** .5  # gradient magnitude
N = G / G.max()  # normalize 0..1

plt3d.plot_surface(xx, yy, z, rstride=1, cstride=1,
                    facecolors=cm.jet(N),
                    linewidth=0, antialiased=False, shade=False,
                    )
plt.show()



"""
    Peaks.EghTauLimits.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from .ElutionModels import egh
from SecTheory.BasicModels import robust_single_pore_pdf

# t = k*(x - t0)
# Ksec = (1 - rho)**

def get_stochastic_model(x, y):

    def objective(p):
        t0, h, k, np_, tp_ = p
        return np.sum((h*robust_single_pore_pdf(k*(x - t0), np_, tp_) - y)**2)

    ret = minimize(objective, (0, 30, 1, 150, 1))
    return ret

def demo(taus=np.arange(-10, 20).reshape((6,5))):
    nrows, ncols = taus.shape
    # fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(3*ncols, nrows*2))
    fig = plt.figure(figsize=(3*ncols + nrows*2, nrows*2))
    gs = GridSpec(nrows, ncols + (ncols-1))

    fig.suptitle("Single Pore Model Correspondence to EGH Model where sigma=20", fontsize=30)

    x = np.arange(300)
    fv_s = []
    for i in range(nrows):
        for j in range(ncols):
            tau = taus[i,j]
            ax = fig.add_subplot(gs[i,j])
            ax.set_title("tau=%g" % tau, fontsize=16)
            y = egh(x, 1, 150, 20, tau)
            ax.plot(x, y, label="egh")
            ret = get_stochastic_model(x, y)
            print("ret.x=", ret.x)
            fv_s.append(ret.fun)
            t0, h, k, np_, tp_ = ret.x
            y_ = h*robust_single_pore_pdf(k*(x - t0), np_, tp_)
            ax.plot(x, y_, label="stochastic")
            ax.legend()

    ax = fig.add_subplot(gs[:,ncols:])
    ax.set_title("EGH Deviations from Stochastic Model", fontsize=20)
    ax.set_xlabel("Tau", fontsize=16)
    ax.set_ylabel("Deviation", fontsize=16)
    ax.plot(taus.flatten(), fv_s, "o-")

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)
    plt.show()



"""
    Models/Stochastic/TriporeImpl.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from matplotlib.widgets import Slider, Button
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.QuickAnalysis.ModeledPeaks import recognize_peaks
from molass_legacy.Models.ElutionCurveModels import egha
from molass_legacy.Models.Tentative import Model
from molass_legacy.Models.ElutionModelUtils import x_from_height_ratio_impl
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy.SecTheory.SecCF import gec_tripore_phi
from molass_legacy.SecTheory.SecPDF import FftInvPdf

PORESIZES = np.array([170,  55,  40])   # see meeting-docs/20231218/index.html
POREPROPS = np.array([0.1, 0.3, 0.6])   # see as above
# N = 1000
# T = 1.2

tripore_pdf = FftInvPdf(gec_tripore_phi)

def tripore_func(x, scale, *params):
    """
    all: N, T, me, mp, pore-distribution, Rg, scale, x0
    pore-distribution:
        pore sizes
        pore proportions

    mid: N, T, Rg, scale, x0  ---- implemented below

    min: Rg, scale, x0
    """

    N, T, Rg, x0 = params

    me = 1.5
    mp = 1.5
    rho = Rg/PORESIZES
    rho[rho > 1] = 1
    np_ = N * POREPROPS * (1 - rho)**me
    tp_ = T*(1 - rho)**mp

    return scale*tripore_pdf(x, np_[0], tp_[0], np_[1], tp_[1], np_[2], tp_[2], x0)

VERY_SMALL_VALUE = 1e-8
if False:
    save_reg_data_fh = open("reg-data.csv", "w")
else:
    save_reg_data_fh = None

def guess_init_params_better(x, y, M):
    area = np.sum(y)
    score_list = []
    params_list = []
    for uscale in [3]:
        t0 = M[1]/2
        u = 30/M[1]*uscale
        a = 0.5
        if M[3] < 0:
            b = -4.0
        else:
            b = -4.0
        e = 0.4
        Dz = 0.02
        cinj = M[0]/2.0 * 0.2
        params = np.array([t0, u, a, b, e, Dz, cinj])
        params_list.append(params)
        y_ = tripore_func(x, *params)
        area_ = np.sum(y_)
        score = abs(1 - area_/area)

        if np.isnan(score):
            score = np.inf

        score_list.append(score)

    k = np.argmin(score_list)
    return params_list[k]

def guess_init_params(x, y, mu, M, rg, debug=False):
    N = 2000
    T = 8
    x0 = 50
    def objective(p, title=None):
        scale, N, T, rg, x0 = p
        y_ = tripore_func(x, scale, N, T, rg, x0)
        if title is not None:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(title)
                ax.plot(x, y)
                ax.plot(x, y_)
                fig.tight_layout()
                plt.show()
        return np.sum((y_ - y)**2)

    print("mu/M[1]=", mu/M[1])
    init_params = (1, N, T, rg, x0)
    if debug:
        objective(init_params, title="guess_init_params: before minimize")

    ret = minimize(objective, init_params)

    if debug:
        objective(ret.x, title="guess_init_params: after minimize")
    scale_, N_, T_, rg_, x0_ =  ret.x

    return np.array([scale_, N_, T_, rg_, x0_])

def guess(x, y, mu, init_params=None, debug=False, debug_info=None):
    from molass_legacy.Models.ElutionModelUtils import compute_4moments

    if debug:
        # y = tripore_func(x, scale, np1, tp1, np2, tp2, np3, tp3, x0)
        def debug_plot_with_sliders():
            slider_specs = [    ("scale", 0, 10, init_params[0]),
                                ("N", 0, 2000, init_params[1]),
                                ("T", 0, 10, init_params[2]),
                                ("rg", 0, 100, init_params[3]),
                                ("x0",  -100, 100, init_params[4]),
                                ]

            with plt.Dp():
                fig, ax = plt.subplots()
                dp = plt.get_dp()

                ax.set_title("guess: before minimize with sliders")
                ax.plot(x, y)

                stc_line, = ax.plot(x, tripore_func(x, *init_params))

                def slider_update(k, val):
                    # print([k], "slider_update", val)
                    init_params[k] = val
                    y_ = tripore_func(x, *init_params)
                    stc_line.set_data(x, y_)
                    dp.draw()

                slider_axes = []
                sliders = []
                for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
                    ax_ = fig.add_axes([0.15, 0.8 - 0.08*k, 0.25, 0.03])
                    slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
                    slider.on_changed(lambda val, k_=k: slider_update(k_, val))
                    slider_axes.append(ax_)
                    sliders.append(slider)

                def reset_slider_valules(event):
                    for slider in sliders:
                        slider.reset()
                    dp.draw()

                axreset = fig.add_axes([0.9, 0.0, 0.1, 0.075])
                breset = Button(axreset, 'Reset')
                breset.on_clicked(reset_slider_valules)

                fig.tight_layout()
                applied = plt.show()
                if applied:
                    params = np.array([slider.val for slider in sliders])
                else:
                    params = init_params

            return params

        def debug_plot_params(x, y, params, title):
            print("params=", params)
            wx, wy, wy_resid, range_ = debug_info
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2,figsize=(12,5))
                fig.suptitle(title, fontsize=20)
                ax1.set_title("Guessing Range", fontsize=16)
                ax2.set_title("Guess in the Range", fontsize=16)

                ax1.plot(wx, wy)
                ax1.plot(wx, wy_resid)
                ax1.plot(x, y)
                ymin, ymax = ax1.get_ylim()
                ax1.set_ylim(ymin, ymax)
                for px in x[[0,-1]]:
                    ax1.plot([px, px], [ymin, ymax], ":", color="gray")

                ax2.plot(x, y)
                ax2.plot(x, tripore_func(x, *params))
                fig.tight_layout()
                plt.show()

    if init_params is None:
        M = compute_4moments(x, y)
        # init_params = guess_init_params_better(x, y, M)
        init_params = guess_init_params(x, y, mu, M, 60)
        area = np.sum(y)
        y_i = tripore_func(x, *init_params)
        area_i = np.sum(y_i)
        ratio = area_i/area
        print("area ratio=", ratio)
        if debug:
            print("M=", M)
            print("init_params=", init_params)
            changed_params = debug_plot_with_sliders()
            print("changed_params=", changed_params)
            changed_y_i = tripore_func(x, *changed_params)
            changed_area_i = np.sum(changed_y_i)
            changed_ratio = changed_area_i/area
            print("changed area ratio=", changed_ratio)
            if ratio < 0.1:
                print("changed_params=", changed_params)
                init_params = changed_params
            debug_plot_params(x, y, init_params, "guess: before minimize")

    def objective(p):
        y_ = tripore_func(x, *p)
        return np.sum((y_ - y)**2)

    ret = minimize(objective, init_params)

    if debug:
        print("M=", M)
        debug_plot_params(x, y, ret.x, "guess: after minimize")

    return ret.x

def guess_multiple_impl(x, y, num_peaks, debug=False):
    egha_params_array = np.array(recognize_peaks(x, y, exact_num_peaks=num_peaks, affine=True))

    cy_list = []
    for params in egha_params_array:
        cy = egha(x, *params)
        cy_list.append(cy)

    if debug:
        # np.savetxt("temp.dat", np.array([x, y]).T)
        ty = np.sum(cy_list, axis=0)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("guess_multiple: EGH decomposition")
            ax.plot(x, y)
            for cy in cy_list:
                ax.plot(x, cy, ":")
            ax.plot(x, ty, ":", color="red", lw=3)
            fig.tight_layout()
            plt.show()

    # sort in the order of peak heights
    exec_recs = sorted(np.array([np.arange(num_peaks), egha_params_array[:,0]]).T, key=lambda x: -x[1])
    print("exec_recs=", exec_recs)

    ecurve = Struct(x=x, y=y)   # dummy will sffice here
    stc_params_list = [None] * num_peaks
    y_resid = y.copy()
    edm_cy_list = [None] * num_peaks

    num_failures = 0
    for k, h_ in exec_recs:
        i = int(k)
        egha_params = egha_params_array[i,:]
        try:
            ret_xes = x_from_height_ratio_impl(egha, ecurve, 0.1, *egha_params[1:])
            # print([i], egha_params, ret_xes)
        except:
            # as with pH6
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "a possible component peak ignored due to x_from_height_ratio_impl failure: ")
            num_failures += 1
            continue

        range_ = np.logical_and(x > ret_xes[0], x < ret_xes[1])
        x_ = x[range_]
        y_ = y_resid[range_]
        if debug:
            debug_info = (x, y, y_resid, range_)
        else:
            debug_info = None

        if np.sum(y_) < 0:
            # avoid computing for negative regions
            num_failures += 1
            continue

        mu = egha_params[1]
        params = guess(x_, y_, mu, debug=debug, debug_info=debug_info)
        print([i], "params=", params)
        stc_params_list[i] = params
        cy = tripore_func(x, *params)
        edm_cy_list[i] = cy
        y_resid -= cy

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("guess_multiple: STC decomposition")
            ax.plot(x, y)

            cy_list = []
            for cy in edm_cy_list:
                if cy is not None:
                    ax.plot(x, cy, ":")
                    cy_list.append(cy)

            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red", lw=3)

            fig.tight_layout()
            plt.show()

    if num_failures > 0:
        ret_list = []
        for p in stc_params_list:
            if p is not None:
                ret_list.append(p)
    else:
        ret_list = stc_params_list

    return np.array(ret_list)


"""
    SpecialSigmoid.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt
from .Sigmoid import fit_bent_sigmoid, ex_sigmoid

def guess_special_sigmoid_params(fc, debug=False):
    ecurve = fc.a_curve

    assert len(ecurve.peak_info) == 1

    ecurve2 = fc.a_curve2
    x = ecurve.x
    y = ecurve.y
    x2 = ecurve2.x
    y2 = ecurve2.y

    f, t = ecurve.get_peak_region_sigma()

    y_ = y.copy()
    y_[np.logical_and(x > f, x < t)] = 0      # better 

    px = (f + t)/2
    popt, pcov = fit_bent_sigmoid(x, y_, x0=px)

    if debug:
        from molass_legacy.Elution.CurveUtils import simple_plot
        with plt.Dp():
            fig, (ax1,ax2) = plt.subplots(nrows=2, figsize=(6,8))
            fig.suptitle("guess_special_sigmoid_params: debug")

            simple_plot(ax1, fc.a_curve)
            ymin, ymax = ax1.get_ylim()

            p = Rectangle(
                    (f, ymin),      # (x,y)
                    t - f,          # width
                    ymax - ymin,    # height
                    facecolor   = 'cyan',
                    alpha       = 0.2,
                )

            ax1.add_patch(p)

            ax1.plot(x, ex_sigmoid(x, *popt))

            ax2.plot(x2, y2)

            fig.tight_layout()
            plt.show()

    return [popt]



# coding: utf-8
"""
    ScatteringPlot.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
import copy
import re
from bisect                 import bisect_right
import matplotlib.pyplot    as plt
import matplotlib.gridspec  as gridspec
from mpl_toolkits.mplot3d   import Axes3D
from molass_legacy._MOLASS.SerialSettings         import get_setting, get_xray_picking
from molass_legacy.KekLib.OurTkinter             import Tk
from molass_legacy.KekLib.TkUtils                import split_geometry
from ResultGui              import ResultGui
from ScatteringBaseUtil     import apply_baseline_correction_impl
from ScatteringBasesurface  import ScatteringBasesurface, is_to_plot

DEBUG           = False
ZLIM_EXPAND     = 1.25

class ScatteringPlot:
    def __init__( self, qvector, curve_y, intensity_array, xray_slice, affine_info, parent=None ):
        self.qvector    = qvector
        self.index      = bisect_right( qvector, get_xray_picking() )
        self.curve_y    = curve_y
        self.data       = intensity_array
        self.xray_slice = xray_slice
        self.affine_info    = affine_info
        self.parent     = parent
        self.basesurfaces = []
        self.ax2        = None
        self.ax3        = None
        self.very_small_angle_only  = False

    def draw_3d_scattering( self, ax, title, data, curve_y, zlim=None, zlim_expand=None ):

        ax.set_title( title )
        ax.set_xlabel( '\nQ(Å⁻¹)' )
        ax.set_ylabel( '\nsequential №' )
        ax.set_zlabel( '\nintensity' )
        # ax.zaxis._set_scale('log')

        if zlim is None:
            if zlim_expand is None:
                zlim_expand = ZLIM_EXPAND
            zmin    = np.min( self.curve_y )
            zmax    = np.max( self.curve_y )
            # zmin, zmax = ax.get_zlim()
            zmin_   = zlim_expand * zmin + ( 1-zlim_expand ) * zmax
            zmax_   = ( 1-zlim_expand ) * zmin + zlim_expand * zmax
            zlim    = ( zmin_, zmax_ )

        ax.set_zlim( zlim )

        print( 'data.shape=', data.shape )

        size = data.shape[0]

        peak_j = None
        for i, q in enumerate( self.qvector ):
            if not is_to_plot( self, i, q, very_small_angle_only=self.very_small_angle_only ):
                continue

            X = np.ones( size ) * q
            Y = np.arange( size )

            if i == self.index:
                alpha   = 1
                color   = 'orange'
                Z = curve_y
            else:
                alpha   = 0.2
                color   = '#1f77b4'
                Z = data[:,i,1]

            ax.plot( X, Y, Z, color=color, alpha=alpha )

    def draw_3d_for_gui( self, parent, title, do_correction, baseline_degree ):
        self.left_title  = title
        self.parent = parent
        from DebugCanvas            import DebugCanvas
        from molass_legacy.KekLib.TkUtils                import is_low_resolution

        def draw_func( fig ):
            # fig.set_size_inches( 20, 10 )
            gs  = gridspec.GridSpec( 1, 2 )
            ax1 = fig.add_subplot( gs[0,0], projection='3d' )
            # ax2 = fig.add_subplot( gs[0,1] )
            self.fig    = fig
            self.gs     = gs
            self.ax1    = ax1
            # self.ax2    = ax2
            self.draw_3d_scattering( ax1, title, self.data, self.curve_y )
            # self.draw_3d_scatterring_detail( ax1, ax2 )
            if do_correction:
                self.add_basesurface( self.data )
                self.basesurfaces[0].plot( self.ax1 )
            fig.tight_layout()

        figsize = ( 16, 8 ) if is_low_resolution() else ( 20, 10 )
        self.canvas = DebugCanvas( "Scattering Data 3D Plot", draw_func,
                        parent=parent, figsize=figsize )

        def connect_popup():
            # this function must be called after self.canvas.show
            self.create_popup_menu()
            # pick_event does not seem to work in this case
            self.canvas.mpl_canvas.mpl_connect( 'button_press_event', self.popup )

        self.parent.after( 100, connect_popup )
        self.do_correction  = do_correction
        self.baseline_degree = baseline_degree
        self.canvas.show( cunstom_button_cb=self.cunstom_button_box, cursor_update=False )

    def create_popup_menu( self ):
        self.popup_menu = Tk.Menu( self.canvas, tearoff=0 )
        self.popup_menu.add_command( label='Guinier Analysis',     command=self.guinier_analysis )
        self.popup_menu_count = 1

    def popup( self, event ):
        if event.button != 3:
            return

        self.popup_event = event
        # print( 'popup: x=%g, y=%g' % ( event.x, event.y ) )

        self.file_index = self.get_file_index( event )
        if self.file_index is None:
            return

        base_widget = self.canvas
        w, h, x, y = split_geometry( base_widget.mpl_canvas_widget.winfo_geometry() )
        h_ = h
        # print( 'self.mpl_canvas_widget.winfo_geometry()=', (w, h, x, y) )
        w, h, x, y = split_geometry( base_widget.winfo_geometry() )
        # print( 'self.winfo_geometry()=', (w, h, x, y) )
        # TODO: better way to get the mouse cursor position
        file_index_str = str(self.file_index)
        print( 'file_index_str=', file_index_str )
        self.popup_menu.entryconfig( 0, label="Guinier Analysis on y=" + file_index_str )
        if self.popup_menu_count == 2:
            self.popup_menu.entryconfig( 1, label="Guinier Analysis for both on y=" + file_index_str )
        self.popup_menu.post( int( x + event.x + 10 ), int( y + h_ - event.y + 30 ) )

    def get_file_index( self, event ):
        if event.ydata is None:
            return None

        # print( "guinier_analysis", (event.x, event.y) )
        # print( "guinier_analysis", (event.xdata, event.ydata) )
        """
            learned from stackoverflow article
            https://stackoverflow.com/questions/6748184/matplotlib-plot-surface-get-the-x-y-z-values-written-in-the-bottom-right-cor
        """

        y_value_re = re.compile( r'y=(-?\d+\.?\d*)' )

        ax = event.inaxes
        if ax == self.ax1:
            print( 'ax1' )
            self.current_data   = self.data
        elif ax == self.ax3:
            print( 'ax3' )
            self.current_data   = self.corrected_data
        else:
            print( 'ax2' )
            return None

        coord3d = ax.format_coord( event.xdata, event.ydata )
        print( 'coord3d=', coord3d )

        m   = y_value_re.search( coord3d )
        if not m:   return None

        file_index = int( float( m.group(1) ) )
        if file_index >= 0 and file_index < self.data.shape[0]:
            return file_index
        else:
            return None

    def cunstom_button_box( self, frame ):
        box1 = Tk.Frame( frame )
        box1.pack( fill=Tk.X, anchor=Tk.CENTER )
        self.button_frame = Tk.Frame( box1 )
        self.button_frame.grid( row=0, column=0 )
        box1.columnconfigure( 0, weight=1 )
        # box1.columnconfigure( 1, weight=1 )
        box2 = Tk.Frame( frame )
        box2.pack( fill=Tk.X )

        w = Tk.Button(box2, text="Close", width=10, command=self.canvas.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.RIGHT, padx=10, pady=10)
        self.ok_button = w

        self.cunstom_button_box_add()

        self.canvas.bind("<Return>", self.canvas.ok)

    def add_basesurface( self, data ):
        print( 'add_basesurface' )
        basesurface = ScatteringBasesurface( self.qvector, self.index, data,
                                inty_curve_y=self.curve_y,
                                parent=self.parent, baseline_degree=self.baseline_degree )

        self.basesurfaces.append( basesurface )

    def cunstom_button_box_add( self ):
        self.entire = Tk.IntVar()
        self.entire.set( 1 )
        cb = Tk.Checkbutton( self.button_frame, text="Entire Scattering", variable=self.entire, state=Tk.DISABLED )
        cb.grid( row=0, column=0, padx=5 )

        self.base = Tk.IntVar()
        self.base.set( 1 )
        cb = Tk.Checkbutton( self.button_frame, text="Base Surface", variable=self.base, state=Tk.DISABLED )
        cb.grid( row=0, column=1, padx=5 )

        self.vsa_only = Tk.IntVar()
        self.vsa_only.set( 0 )
        self.vsa_only_cb = Tk.Checkbutton( self.button_frame, text="Very Small Angle only", variable=self.vsa_only )
        self.vsa_only_cb.grid( row=0, column=2, padx=5 )

        space = Tk.Frame( self.button_frame, width=10 )
        space.grid( row=0, column=3, padx=10 )

        label = Tk.Label( self.button_frame, text="Zlim expand ratio:" )
        label.grid( row=0, column=4 )
        self.zlim_expand = Tk.DoubleVar()
        self.zlim_expand.set( 1.25 )
        entry = Tk.Entry( self.button_frame, textvariable=self.zlim_expand, width=6, justify=Tk.RIGHT )
        entry.grid( row=0, column=5 )

        space = Tk.Frame( self.button_frame, width=10 )
        space.grid( row=0, column=6, padx=10 )

        self.redraw = Tk.Button( self.button_frame, text="Redraw", command=self.redraw )
        self.redraw.grid( row=0, column=7, padx=10 )

        view_init = Tk.Button( self.button_frame, text="View Init", command=self.view_init )
        view_init.grid( row=0, column=8, padx=10 )

        self.change_to_corrected_3d()
        self.canvas.update()

    def redraw( self ):
        self.canvas.config( cursor='wait' )
        self.canvas.update()

        self.very_small_angle_only = self.vsa_only.get() == 1

        zlim_expand = self.zlim_expand.get()
        self.ax1.cla()
        self.draw_3d_scattering( self.ax1, self.left_title, self.data, self.curve_y, zlim_expand=zlim_expand )
        self.basesurfaces[0].plot( self.ax1, very_small_angle_only=self.very_small_angle_only )

        self.ax3.cla()
        self.draw_3d_scattering( self.ax3, self.right_title, self.corrected_data, self.corrected_curve_y, zlim_expand=zlim_expand )
        self.basesurfaces[1].plot( self.ax3, very_small_angle_only=self.very_small_angle_only )
        self.canvas.config( cursor='' )
        self.canvas.mpl_canvas.draw()

    def view_init( self ):
        for ax in [self.ax1, self.ax3]:
            ax.view_init()
        self.canvas.mpl_canvas.draw()

    def guinier_analysis( self, both=False ):
        in_folder   = get_setting( 'in_folder' )
        serial_data = self.parent.serial_data       # only for filename and serial_data.q_limit_index
        filename    = serial_data.datafiles[self.file_index].split( '\\' )[-1]
        print( 'filename=', filename )

        if both:
            data_list   = [ self.data, self.corrected_data ]
        else:
            data_list   = [ self.current_data ]

        result_array = []
        for data in data_list:
            np_array    = data[self.file_index,:,:]

            if data is self.data:
                filename_       = filename
                atsas_np_array  = None
            else:
                filename_       = filename + '(corrected)'
                atsas_np_array  = np_array
            result_gui  = ResultGui( self.canvas, in_folder, filename,
                                        np_array=np_array, serial_data=serial_data,
                                        title_filename=filename_,
                                        atsas_np_array=atsas_np_array )
            result_array.append( result_gui )

    def create_corrected_data( self ):
        jvector = np.arange( self.data.shape[0] )
        qvector = self.data[0,:,0]
        self.corrected_data = copy.deepcopy( self.data )

        print( 'create_corrected_data: baseline_degree=', self.baseline_degree )

        try:
            assert False
            # incompatible with revised apply_baseline_correction_impl
            apply_baseline_correction_impl( jvector, qvector, self.corrected_data,
                                                affine_info=self.affine_info,
                                                inty_curve_y=self.curve_y,
                                                baseline_opt=1,     # force this even in case of 'No correction'
                                                baseline_degree=self.baseline_degree,
                                                parent=self.parent
                                                )
        except RuntimeError as exc:
            # this is known to be harmless
            # because it seemts to occur at the (widest angle) end point.
            pass

        self.corrected_curve_y = np.average( self.corrected_data[:, self.xray_slice, 1], axis=1 )

    def change_to_corrected_3d( self ):
        self.canvas.config( cursor='wait' )
        self.canvas.update()

        self.create_corrected_data()

        if self.ax2 is not None:
            self.toolbar.destroy()
            self.change_button.pack_forget()

        if self.ax3 is None:
            self.ax3 = self.fig.add_subplot( self.gs[0,1], projection='3d' )

        self.right_title = "Corrected scattering made by subtracting the basesurface on the left figure"
        self.draw_3d_scattering( self.ax3, self.right_title, self.corrected_data, self.corrected_curve_y )

        if self.do_correction:
            self.add_basesurface( self.corrected_data )
            self.basesurfaces[1].plot( self.ax3 )

        self.canvas.mpl_canvas.draw()

        self.canvas.config( cursor='' )
        self.canvas.update()
        self.popup_menu.add_command( label='Guinier Analysis for both',     command=lambda: self.guinier_analysis( both=True ) )
        self.popup_menu_count = 2



"""
    Optimizer.JobStateInfo.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import logging
import os
import numpy as np
from datetime import datetime

class JobStateInfo:
    def __init__(self, optinit_info, demo_index):
        self.logger = logging.getLogger(__name__)
        self.optinit_info = optinit_info
        self.dsets = optinit_info.dsets
        self.fullopt = optinit_info.fullopt
        self.demo_index = demo_index
        self.optlogfile = None

        if demo_index is not None:
            if demo_index == 0:
                # for run
                self.prepare_init_state()
            else:
                # for demo
                self.prepare_for_demo()
            self.curr_fv = self.demo_info[0]

    def get_ready_from_folder(self, work_folder):
        self.demo_index = -1
        self.dsets = self.get_dsets_from_work_folder(work_folder)
        self.prepare_init_state()
        self.curr_fv = self.demo_info[0]

    def get_dsets_from_work_folder(self, work_folder):
        from .TheUtils import get_work_info
        from .FullOptInput import FullOptInput
        # for demo to revive
        work_info = get_work_info(work_folder)
        fullopt_input = FullOptInput(in_folder=work_info.in_folder)
        return fullopt_input.get_dsets()

    def get_demo_info(self, work_folder):
        from .StateSequence import StateSequence
        sseq = StateSequence(work_folder=work_folder)
        return sseq.get_info()

    def prepare_for_demo(self):
        from .TheUtils import get_work_info_for_demo
        work_info = get_work_info_for_demo()
        self.logger.info("prepare_for_demo: real_bounds=%s" % work_info.real_bounds)
        self.set_init_info(init_params=work_info.init_params, real_bounds=work_info.real_bounds)
        self.demo_info = self.get_demo_info(work_info.work_folder)

    def get_optlogfile(self):
        if self.optlogfile is None:
            from molass_legacy._MOLASS.SerialSettings import get_setting
            from molass_legacy.Optimizer.OptLogFile import OptLogFile
            work_folder = get_setting("optworking_folder")
            path = os.path.join(work_folder, 'optimizer.log')
            self.optlogfile = OptLogFile(path)
        return self.optlogfile

    def get_active_indeces(self):
        optlogfile = self.get_optlogfile()
        return optlogfile.get_active_indeces()

    def prepare_init_state(self, init_params=None, real_bounds=None):
        self.set_init_info(init_params=init_params, real_bounds=real_bounds)
        self.demo_info = self.get_init_info()

    def set_init_info(self, init_params=None, real_bounds=None):
        if init_params is None:
            init_params = self.optinit_info.init_params

        self.fullopt.prepare_for_optimization(init_params, real_bounds=real_bounds)
        fv = self.fullopt.objective_func(init_params)
        fv_array = np.array([(0, fv, False, datetime.now())])   # should be compatible with .StateSequence.read_callback_txt_impl
        x_array = np.array([init_params])
        self.init_info = fv_array, x_array, 500000, 0

    def get_init_info(self):
        return self.init_info

    def set_demo_info(self, demo_info):
        self.demo_info = demo_info
        self.curr_fv = self.demo_info[0]
        self.draw_suptitle()
        i = self.get_best_index()
        self.best_index = i
        self.curr_index = i
        self.draw_state_impl(i)
        self.draw_progress()



"""
    Selective.Proportions.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np

PROPORTIONS_FUNC_SPEC = [
    [("def proportions(p):\n"
     "    return p, 1-p"), (0.05, 0.2)],
    [("def proportions(p):\n"
     "    return p, 1-2*p, p"), (0.025, 0.15)],
    [("def proportions(p):\n"
     "    return p, 0.5-p, 0.5-p, p"), (0.025, 0.1)],
    [("def proportions(p):\n"
     "    return 2*p, 0.3-p, 0.4-2*p, 0.3-p, 2*p"), (0.01, 0.05)],
    [("def proportions(p):\n"
     "    return 2*p, 0.25-p, 0.25-p, 0.25-p, 0.25-p, 2*p"), (0.01, 0.05)],
    [("def proportions(p):\n"
     "    return 3*p, 0.15-p, 0.15-p, 0.4-2*p, 0.15-p, 0.15-p, 3*p"), (0.007, 0.04)],
]

class Proportions_:         # renamed so as not to be confused with "proportions" function
    def __init__(self):
        pass

    def generate_funccode(self, num_components):
        return PROPORTIONS_FUNC_SPEC[num_components-2][0]

    def get_min_max_values(self, n):
        """
        Returns a tuple of (min, max) values for the proportions function parameter p.
        """
        return PROPORTIONS_FUNC_SPEC[n-2][1]

    def guess_optimal_prop(self, n):
        minv, maxv = self.get_min_max_values(n)
        return minv*0.7 + maxv*0.3

    def get_optimal_props_text(self, n):
        return " proportions(%.3g)" % self.guess_optimal_prop(n)

    def get_with_props_values(self, n):
        # the eval below must be preceeded by the "proportions" function definition,
        # which can be attained by calling self.get_proportions_func(...)
        opt_props_text = self.get_optimal_props_text(n)
        return [opt_props_text, str(eval(opt_props_text))]
    
    def get_proportions_func(self, code_text, debug=False):
        if debug:
            print(code_text)
        context = {}
        exec(code_text, context)
        # note this code_text must define a function named "proportions"
        proportions = context['proportions']
        globals()['proportions'] = proportions      # "proportions" will be called in the eval above
        if debug:
            for p in np.linspace(0.05, 0.2, 20):
                print(proportions(p))
        return proportions



# coding: utf-8
"""
    Rgg.GmmSpike.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time
import asyncio
import numpy as np
from matplotlib.gridspec import GridSpec
from scipy.stats import multivariate_normal
from sklearn import mixture
import MatplotlibUtils          # required for annotate3D
from molass_legacy.Peaks.ElutionModels import egh
from Prob.GaussianMixture import gaussian_pdf
import molass_legacy.KekLib.DebugPlot as plt
from RgProcess.RgCurve import make_availability_slices
from .SpikeData import generate_demo_data, FULL_PARAMS, normal_pdf
from .RggUtils import convert_to_probabilitic_data, plot_histogram_2d

def data_demo():
    x, y, rg, num_components = generate_demo_data()
    max_y = np.max(y)
    slices, states = make_availability_slices(y, max_y=max_y, min_ratio=0.03)

    plt.push()
    fig = plt.figure(figsize=(21,6))
    ax1 = fig.add_subplot(131)
    ax2 = fig.add_subplot(132, projection='3d')
    ax3 = fig.add_subplot(133, projection='3d')
    axt = ax1.twinx()
    axt.grid(False)

    fig.suptitle("Conversion to 2D Probabilistic Data compared to the True Distribution", fontsize=20)

    ax1.set_title("Observed Elution Curve and Rg", fontsize=16)
    ax2.set_title("Apparent Histogram of 2D Data Distribution", fontsize=16)
    ax3.set_title("True Histogram of 2D Data Distribution", fontsize=16)

    for ax in [ax2, ax3]:
        ax.set_xlabel("Eno")
        ax.set_ylabel("Rg")
        ax.set_zlabel("Counts")
        ax.set_ylim(10, 50)

    axt.set_ylim(10, 50)

    for slice_, state in zip(slices, states):
        x_ = x[slice_]
        y_ = y[slice_]
        rg_ = rg[slice_]
        linestyle = '-' if state else ':'
        ax1.plot(x_, y_, linestyle, color='C0')
        axt.plot(x_, rg_, linestyle, color='C1')
        if state:
            plot_histogram_2d(ax2, x_, y_, rg_, max_y)

    peaktop = (200, 35, 100)
    for ax in [ax2, ax3]:
        ax.annotate3D('(%.3g, %.3g, %.3g)' % peaktop, peaktop,
          xytext=(30,-30),
          textcoords='offset points',
          bbox=dict(boxstyle="round", fc="lightyellow"),
          arrowprops = dict(arrowstyle="-|>", ec='black', fc='white', lw=1))

    for rg, params in FULL_PARAMS:
        h, mu, sigma, tau = params
        y_ = h * normal_pdf(x, mu, sigma)
        rg_ = np.ones(len(x))*rg
        plot_histogram_2d(ax3, x, y_, rg_, max_y)

    ax2.set_xlim(ax3.get_xlim())

    fig.tight_layout()
    fig.subplots_adjust(top=0.88, left=0.04, wspace=0.12, right=0.96)
    plt.show()
    plt.pop()

USE_VARIATIONAL = True
USE_SCIKIT_LEARN = True

def spike_demo():
    x, y, rg, num_components = generate_demo_data()
    spike_demo_impl([(x, y, rg, None)], num_components=num_components)

def spike_demo_impl(data_list, num_components=None, num_precision=100):

    num_datasets = len(data_list)

    plt.push()
    fig = plt.figure(figsize=(21,7))
    gs = GridSpec(num_datasets, 3)

    for k, (x, y, rg, X) in enumerate(data_list):
        ax1 = fig.add_subplot(gs[k,0])
        ax2 = fig.add_subplot(gs[k,1], projection='3d')
        ax3 = fig.add_subplot(gs[k,2])

        axt = ax1.twinx()
        axt.grid(False)

        # ax.plot(x, y)
        # axt.plot(x, rg, color='C1')
        max_y = np.max(y)
        X_list = []
        slices, states = make_availability_slices(y, max_y=max_y, min_ratio=0.03)
        for slice_, state in zip(slices, states):
            x_ = x[slice_]
            y_ = y[slice_]
            rg_ = rg[slice_]
            if state:
                X_list.append(convert_to_probabilitic_data(x_, y_, rg_, max_y=max_y, num_precision=num_precision))
            linestyle = '-' if state else ':'
            ax1.plot(x_, y_, linestyle, color='C0')
            axt.plot(x_, rg_, linestyle, color='C1')
            if state:
                if True:
                    plot_histogram_2d(ax2, x_, y_, rg_, max_y, num_precision=num_precision)
                else:
                    b_ = np.zeros(len(x_))
                    ax2.bar3d(x_, rg_, b_, 0.5, 0.05, y_/max_y*100, shade=True, edgecolor='green')
                ax2.set_ylim(10, 60)

        ymin, ymax = axt.get_ylim()
        axt.set_ylim(0, ymax*1.2)

        X = np.concatenate(X_list)
        # covariance_type='spherical'

        max_iter = 100
        covariance_type='full'
        # means_init = [(200, 35), (250, 30), (400, 23)]

        if USE_VARIATIONAL:
            if USE_SCIKIT_LEARN:
                gmm = mixture.BayesianGaussianMixture(n_components=num_components, covariance_type=covariance_type, max_iter=max_iter)
                gmm.fit(X)
            else:
                assert False
        else:
            gmm = mixture.GaussianMixture(n_components=num_components, covariance_type=covariance_type, max_iter=max_iter)
            gmm.fit(X)
        print(dir(gmm))
        print("weights_=", gmm.weights_)
        print("means_=", gmm.means_)
        print("covariances_=", gmm.covariances_)
 
        cy_list = []
        ty = np.zeros(len(x))
        for w, m, cv in zip(gmm.weights_, gmm.means_, gmm.covariances_):
            var = cv if covariance_type == 'spherical' else cv[0,0]
            cy = w * gaussian_pdf(x, m[0], np.sqrt(var))
            cy_list.append(cy)
            ty += cy

        scale = max_y/np.max(ty)

        ax3.plot(x, y)
        for cy in cy_list:
            ax3.plot(x, scale*cy, ':')
        ax3.plot(x, scale*ty, ':', color='red')

    fig.tight_layout()
    cont = plt.show()
    plt.pop()
    return gmm

def spike_demo_real(in_folder, num_components=3, correction=True, use_peaks=False):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from RgProcess.RgCurve import RgCurve

    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    if correction:
        sd = sp.get_corrected_sd()
    else:
        sd = sp.get_sd()
    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    rgc = RgCurve(qv, xr_curve, D, E)
    rgc.proof_plot()
    X = rgc.get_probabilistic_data()
    spike_demo_impl([[xr_curve.x, xr_curve.y, None, X]], num_components=3, use_peaks=use_peaks)

async def get_content(n):
    await asyncio.sleep(3)
    return n + 1

async def f(n):
    content = await get_content(n)
    return content

def coroutine_spike():
    loop = asyncio.get_event_loop()
    v = loop.run_until_complete(f(1))
    print(v)



# coding: utf-8
"""
    Quality.py

    Copyright (c) 2016-2017, Masatsuyo Takahashi, KEK-PF
"""
import numpy        as np
from Settings       import get_setting


DEBUG = False

STDEV_SCORE_SCALE   = 10.0

RAW_BASIC_QUALITY   = 0
RAW_FIT_SCORE       = 6
RAW_STDEV_SCORE     = 4

FIT_CONSISTENCY     = 2

def get_factor_weights():
    return get_setting( 'quality_weighting' )

def q_rg_score_func( qrg ):
    if qrg <= 0.3: return 1
    if qrg >= 0.8: return 0
    return ( 0.8 - qrg ) / 0.5

base_part       = 0.4
bonus_part      = 1 - base_part
bonus_rate      = 10
bonus_boundary  = 1 - 1 / bonus_rate    # 0.9

AIC_SCORE_MIN   =  8
AIC_SCORE_MAX   = 10

def fit_score_func( aic ):
    aic_score = np.log( -aic ) if aic < 0 else 0
    aic_ = min( max( AIC_SCORE_MIN,  aic_score ), AIC_SCORE_MAX )
    fit_score = ( aic_ - AIC_SCORE_MIN ) / ( AIC_SCORE_MAX - AIC_SCORE_MIN )
    return fit_score

def fit_consistency_func( fit_Rg, Rg, adopt_ratio=1 ):
    rg_diff_score   = max( 0, 1 - abs( fit_Rg - Rg ) / fit_Rg )

    if rg_diff_score <= bonus_boundary:
        consistency = base_part * rg_diff_score
    else:
        consistency = base_part * rg_diff_score + bonus_part * ( rg_diff_score - bonus_boundary ) * bonus_rate

    return consistency * adopt_ratio

def compute_atsas_fit_consistency( fit_Rg, atsas_Rg, raw_factors ):
    adopt_ratio = np.power( raw_factors[RAW_FIT_SCORE] * raw_factors[RAW_STDEV_SCORE] * raw_factors[RAW_BASIC_QUALITY], 1/4 )
    atsas_fit_consistency   = fit_consistency_func( fit_Rg, atsas_Rg, adopt_ratio=adopt_ratio ) * get_factor_weights()[ FIT_CONSISTENCY ]
    return atsas_fit_consistency

def fit_consistency_with_stdev( fit_Rg, Rg, Rg_stdev ):
    stdev_ratio         = Rg_stdev / Rg
    stdev_score         = np.exp( -stdev_ratio*STDEV_SCORE_SCALE ) if stdev_ratio > 0 else 0

    fit_consistency_pure    = fit_consistency_func( fit_Rg, Rg )
    fit_consistency     = fit_consistency_pure * stdev_score
    return fit_consistency

class Quality:
    def __init__( self, fit, Rg, Rg_stdev, interval, positive_ratio, x ):

        # self.fit    = fit         # this is bad for garbage collection
        self.fit_Rg = fit.Rg

        self.Rg     = Rg
        basic_condition     = fit.intensity.basic_condition
        basic_quality       = fit.intensity.basic_quality
        positive_score      = 2 * ( max( 0.5, positive_ratio ) - 0.5 )

        fit_score   = fit_score_func( fit.result.aic )

        if Rg is None:
            fit_consistency_pure    = 0
            fit_consistency     = 0
            stdev_ratio         = 1
            stdev_score         = 0
            q_rg_score          = 0
        else:
            stdev_ratio         = Rg_stdev / Rg
            if DEBUG: print( 'stdev_ratio=', stdev_ratio )
            stdev_score         = np.exp( -stdev_ratio*STDEV_SCORE_SCALE ) if stdev_ratio > 0 else 0

            fit_consistency_pure= fit_consistency_func( fit.Rg, Rg )
            fit_consistency     = fit_consistency_pure * np.power( fit_score * stdev_score * basic_quality, 1/4 )

            f2_, t2_ = interval
            # size_score      = min( 20, t2_ - f2_ + 1 ) / 20
            q_rg_score = q_rg_score_func( np.sqrt( x[f2_] ) * Rg )
            q_rg_score *= basic_quality

        if DEBUG:
            print( 'fit.cover_ratio=', fit.cover_ratio )
            print( 'fit_score=', fit_score )
            print( 'fit_consistency=', fit_consistency )
            print( 'stdev_score=', stdev_score )
            print( 'q_rg_score=', q_rg_score )

        self.aic_score          = np.log( -fit.result.aic )     # internal use only
        self.basic_condition    = basic_condition               # internal use only

        self.basic_quality      = basic_quality
        self.positive_score     = positive_score
        self.fit_cover_ratio    = fit.cover_ratio * basic_quality
        self.fit_score          = fit_score
        self.fit_consistency_pure   = fit_consistency_pure
        self.fit_consistency    = fit_consistency
        self.stdev_ratio        = stdev_ratio
        self.stdev_score        = stdev_score * basic_quality
        self.q_rg_score         = q_rg_score
        self.needs_bico_solver   = False

        self.update_quality()

    def update_quality( self ):
        self.quality =  np.sum( self.get_factors() )

        if self.Rg is None:
            self.quality = max( 0, self.quality - 0.1 )

        if self.quality <= 0.3:
            self.quality_signal = 'R'
        elif self.quality >= 0.7:
            self.quality_signal = 'G'
        else:
            self.quality_signal = 'Y'

    def get_raw_factors( self ):
        return  [   self.basic_quality,
                    self.positive_score,
                    self.fit_cover_ratio,
                    self.fit_consistency,
                    self.stdev_score,
                    self.q_rg_score,
                    self.fit_score,
                ]

    def get_factors( self ):
        weights = get_factor_weights()
        return  [   weights[0] * self.basic_quality,
                    weights[1] * self.positive_score,
                    0 * self.fit_cover_ratio,
                    weights[2] * self.fit_consistency,
                    weights[3] * self.stdev_score,
                    weights[4] * self.q_rg_score,
                ]

    def get_factors_with_fit_score( self ):
        return self.get_factors() + [ self.fit_score ]

    def degrade_by_anomalies( self, deg_score ):
        self.fit_score = max( 0, self.fit_score - deg_score )
        if self.Rg is not None:
            self.fit_consistency = fit_consistency_func( self.fit_Rg, self.Rg,
                                    adopt_ratio=np.power( self.fit_score * self.stdev_score * self.basic_quality, 1/4 ) )
        self.update_quality()

def is_better_quality( r1, r2, qrg_limits, allow_qrg=0.0 ):
    # print( 'is_better_quality: r1=', r1, ', r2=', r2 )

    if r1 is None:
        return False

    if r2 is None:
        return True

    if r1.Rg is None:
        return False

    if r2.Rg is None:
        return True

    qrg_limit = qrg_limits[1] + allow_qrg
    if r1.max_qRg > qrg_limit:
        if DEBUG: print( 'r1.max_qRg(%g) > qrg_limit(%g)' % ( r1.max_qRg, qrg_limit  ) )
        if r2.max_qRg <= qrg_limit:
            return False

    if r2.max_qRg > qrg_limit:
        if DEBUG: print( 'r2.max_qRg(%g) > qrg_limit(%g)' % ( r2.max_qRg, qrg_limit  ) )
        return True

    if DEBUG:
        print( 'Rg             : %.4g, %.4g' % ( r1.Rg, r2.Rg ) )
        print( 'Aggregated     : %s, %s' % ( r1.Aggregated, r2.Aggregated ) )
        print( 'Quality        : %.4g, %.4g' % ( r1.Quality, r2.Quality ) )
        print( 'basic_quality: %.4g, %.4g' % ( r1.quality_object.basic_quality, r2.quality_object.basic_quality ) )
        print( 'fit_consistency: %.4g, %.4g' % ( r1.quality_object.fit_consistency_pure, r2.quality_object.fit_consistency_pure ) )

    """
    if r1.Aggregated is None:
        if r2.Aggregated is None:
            pass
        else:
            return False
    else:
        if r2.Aggregated is None:
            return True
        else:
            if r1.Aggregated == r2.Aggregated:
                pass
            else:
                # take it if any anomaly is detected
                if abs( r1.Aggregated ) > abs( r2.Aggregated ):
                    return True
    """

    if r1.quality_object.basic_quality > 0.5 and r2.quality_object.fit_consistency_pure > 0.2:
        # print( "is_better_quality fit_consistency_pure:", r1.quality_object.fit_consistency_pure, r2.quality_object.fit_consistency_pure )
        if r1.quality_object.fit_consistency_pure > r2.quality_object.fit_consistency_pure:
            return True
        else:
            return False
    else:
        if r1.Quality > r2.Quality:
            return True
        else:
            return False



"""
    Experiment.ColumnTypes.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np

class ColumnType:
    def __init__(self, **entries): 
        self.__dict__.update(entries)

    def __str__(self):
        return str(self.__dict__)

COLUMN_TYPE_DICT = {
    "ad75w"   : ColumnType( name="Superdex 75 increase 10/300", shortname="Superdex 75-10",
                        size=(10,  300), excl_limit=100, frac_range=(3, 70), num_plates=43000, poresize=35.6),

    "ad75n"   : ColumnType( name="Superdex 75 increase 3.2/300", shortname="Superdex 75-3.2",
                        size=(3.2, 300), excl_limit=100, frac_range=(3, 70), num_plates=43000, poresize=35.6),

    "ad200w"  : ColumnType( name="Superdex 200 increase 10/300", shortname="Superdex 200-10",
                        size=(10,  300), excl_limit=1300, frac_range=(10, 600), num_plates=48000, poresize=76.0),

    "ad200n"  : ColumnType( name="Superdex 200 increase 3.2/300", shortname="Superdex 200-3.2",
                        size=(3.2, 300), excl_limit=1300, frac_range=(10, 600), num_plates=48000, poresize=76.0),

    "ad6w"    : ColumnType( name="Superdex 6 increase 10/300", shortname="Superdex 6-10",
                        size=(3.2, 300), excl_limit=40000, frac_range=(5, 5000), num_plates=48000, poresize=209),

    "nacalai" : ColumnType( name="COSMOSIL 5Diol-300-II (Nacalai)", shortname="COSMOSIL",
                        size=(7.5, 300), excl_limit=135000, frac_range=(10, 700), num_plates=np.nan, poresize=300),

    "agilentn": ColumnType( name="Bio SEC-3 (Φ4.6×300, 300Å) (Agilent)", shortname="Bio SEC-3 (4.6)",
                        size=(4.6, 300), excl_limit=135000, frac_range=(5, 1250), num_plates=np.nan, poresize=300),

    "agilentw": ColumnType( name="Bio SEC-3 (Φ7.8×300, 300Å) (Agilent)", shortname="Bio SEC-3 (7.8)",
                        size=(7.8, 300), excl_limit=135000, frac_range=(5, 1250), num_plates=np.nan, poresize=300),

    "shodex"  : ColumnType( name="KW403-4F（Shodex）", shortname="KW403-4F",
                        size=(4.6, 300), excl_limit=600, frac_range=(10, 500), num_plates=35000, poresize=60.4),

    "unknown" : ColumnType( name="Unknown", shortname="Unknown",
                        size=(np.nan, np.nan), excl_limit=np.nan, frac_range=(np.nan, np.nan), num_plates=np.nan, poresize=np.nan),
}

from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

PORESIZE_ALLOWANCE = 5

names = []
def get_productnames():
    if len(names) == 0:
        for k, v in COLUMN_TYPE_DICT.items():
            names.append(v.name)
    return names

def get_columntype_from_id(id_):
    return COLUMN_TYPE_DICT.get(id_)

def get_columntype_from_name(name, with_id=False):
    coltype = None
    for k, v in COLUMN_TYPE_DICT.items():
        if v.name == name:
            coltype = v
            break
    if with_id:
        return coltype, k
    else:
        return coltype

def get_columntype_id_from_name(name):
    coltype_id = None
    for k, v in COLUMN_TYPE_DICT.items():
        if v.name == name:
            coltype_id = k
            break
    return coltype_id

def update_sec_settings_by_id(columntype_id):
    columntype = COLUMN_TYPE_DICT[columntype_id]

    set_setting('columntype_id', columntype_id)
    set_setting('exclusion_limit', columntype.excl_limit)
    poresize = columntype.poresize
    set_setting('poresize', poresize)
    set_setting('poresize_bounds', (poresize - PORESIZE_ALLOWANCE, poresize + PORESIZE_ALLOWANCE))

def get_columnname():
    columntype_id = get_setting('columntype_id')
    return COLUMN_TYPE_DICT[columntype_id].shortname

def get_all_poresizes():
    poresize_list = []
    for k, v in COLUMN_TYPE_DICT.items():
        poresize_list.append(v.poresize)
    return sorted(list(set(poresize_list)))     # remove duplicates



# coding: utf-8
"""
    ModelDataSet.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt 
from molass_legacy.Peaks.ElutionModels import egh
from Theory.JsPedersen1997 import F1
from molass_legacy.KekLib.SciPyCookbook import smooth
from scipy.interpolate import UnivariateSpline
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from RgProcess.RgCurve import RgCurve

class DataSet:
    def __init__(self, D, E, vec, curve):
        self.D = D
        self.E = E
        self.vec = vec
        self.curve = curve

class UvSpectrum:
    def __init__(self, params=None):
        self.params = params

    def fit(self, wv, y, debug=False):
        sy = smooth(y)
        spline = UnivariateSpline(wv, sy, s=0)
        y1 = spline.derivative(1)(wv)
        y1_spline = UnivariateSpline(wv, y1, s=0)
        roots = y1_spline.roots()
        y2 = spline.derivative(2)(wv)
        y2_spline = UnivariateSpline(wv, y2, s=0)

        max_y = np.max(y)
        wanted = np.logical_and(spline(roots)/max_y > 0.05, y2_spline(roots) < 0)
        wanted_roots = roots[wanted]

        assert len(wanted_roots) >= 2
        sigma = (wanted_roots[1] - wanted_roots[0])/4

        init_params = []
        for r in wanted_roots[0:2]:
            h = spline(r)
            m = r
            s = sigma
            t = 0
            init_params.append((h, m, s, t))

        init_params = np.array(init_params)
        if debug:
            def debug_plot(params):
                plt.push()
                fig, ax = plt.subplots()
                axt = ax.twinx()
                axt.grid(False)

                ax.plot(wv, y)
                for r in wanted_roots:
                    ax.plot(r, spline(r), 'o', color='red')
                axt.plot(wv, y1, ':', color='orange')

                ty = np.zeros(len(wv))
                for (h, m, s, t) in params:
                    cy = egh(wv, h, m, s, t)
                    ty += cy
                    ax.plot(wv, cy, ':')
                ax.plot(wv, ty, ':', color='red')

                plt.show()
                plt.pop()

            # debug_plot(init_params)

        def obj_func(p):
            params = p.reshape(init_params.shape)
            ty = np.zeros(len(wv))
            for (h, m, s, t) in params:
                cy = egh(wv, h, m, s, t)
                ty += cy
            return np.sum((ty - y)**2)

        result = minimize(obj_func, init_params.flatten())
        self.params = result.x.reshape(init_params.shape)

        if debug:
            debug_plot(self.params)

    def eval(self, wv):
        y = np.zeros(len(wv))
        for (h, m, s, t) in self.params:
            y += egh(wv, h, m, s, t)
        return y

FULL_PARAMS =  np.array([
        [35, [1.0, 200, 20,  0]],
        [31, [0.2, 250, 20,  0]],
        [23, [0.8, 400, 30,  0]],
        ])

def scale_params(scale, params):
    params_ = np.array(params)
    params_[1:] *= scale
    return params_

class RgCurveProxy(RgCurve):
    def __init__(self, qv, ecurve, cy_list, min_ratio=0.03):
        from RgProcess.RgCurve import make_availability_slices
        self.x = x = ecurve.x
        self.y = y = ecurve.y
        self.x_ = x
        self.y_ = y
        rg_ = np.zeros(len(self.x))
        max_y = np.max(y)
        slices, states = make_availability_slices(y, max_y=max_y, min_ratio=min_ratio)

        for r, cy in zip(FULL_PARAMS[:,0], cy_list):
            rg_ += r*cy/y

        self.rg_ = rg_
        self.ecurve = ecurve
        self.slices = slices
        self.states = states
        self.segments = self.create_segments(x, y, rg_, slices, states)
        self.X = None

    def create_segments(self, x, y, rg_, slices, states):
        from molass_legacy.KekLib.SciPyCookbook import smooth
        segments = []
        rg_buffer = np.zeros(len(x))
        for slice_, state in zip(slices, states):
            if state == 0:
                continue

            segments.append((x[slice_], y[slice_], rg_[slice_]))
        return segments

    def get_curve_segments(self):
        return self.segments

def generate_model_dsets(scale=1, drift_type=None, debug=False):

    xe_size = int(600*scale)
    xa_size = int(500*scale)
    x = np.arange(xe_size)
    y = np.zeros(len(x))
    qv = np.linspace(0.005, 0.5, xa_size)

    rg_list = []
    cy_list = []
    p_list = []

    def compute_p(rg):
        r = np.sqrt(5/3)*rg
        p = F1(qv, r)**2
        return p

    for rg, params in FULL_PARAMS:
        params_ = scale_params(scale, params)
        cy = egh(x, *params_)
        cy_list.append(cy)
        y += cy
        rg_list.append(rg)
        p = compute_p(rg)
        p_list.append(p)

    if drift_type is None:
        by = np.zeros(len(x))
    else:
        by = x * 0.0002 + 0.01
        rg_ = np.average(FULL_PARAMS[:,0])
        bp = compute_p(rg_)
        p_list += [bp]
        cy_list += [by]

    P = np.array(p_list).T
    C = np.array(cy_list)
    D = P @ C

    (a, b) = (2, 0)
    uv_ends = [x_*a + b for x_ in x[[0,-1]]]
    uv_x = np.linspace(*uv_ends, xe_size*a)
    uv_y = np.zeros(len(uv_x))
    uv_scales = [0.8, 1.0, 1.5]

    uv_cy_list = []
    for uv_scale, (h, m, s, t) in zip(uv_scales, FULL_PARAMS[:,1]):
        m_ = m*scale*a + b
        s_ = s*scale*a
        t_ = t*scale*a
        uv_cy = uv_scale * egh(uv_x, h, m_, s_, t_)
        uv_y += uv_cy
        uv_cy_list.append(uv_cy)

    if drift_type is None:
        uv_by = np.zeros(len(uv_x))
    else:
        uv_by = uv_x * 0.0001 + 0.01
        uv_cy_list += [uv_by]

    UvC = np.array(uv_cy_list)

    #  these params were got from extract_absorption_components("OA_Ald")
    uv_component_params= [
        [[  8.39752249, 218.24631571,  12.30526406,  -4.41740341],
         [  0.9994718,  277.66141429,  14.54344494,  -9.51796811]],
        [[ 10.33094092, 214.09172376,  12.63850912,  -1.00524139],
         [  1.05349782, 271.9127514,   17.12164758,   6.28261841]],
        [[  8.54545972, 219.34075863,  12.410232,    -5.0363978 ],
         [  0.99559913, 276.46754992,  16.58275676, -12.83937896]],
        ]

    uw_size = int(300*scale)
    wv = np.linspace(200, 450, uw_size)

    uv_p_list = []
    for uvs_params in uv_component_params:
        uvs = UvSpectrum(uvs_params)
        uv_p_list.append(uvs.eval(wv))

    if drift_type is None:
        pass
    else:
        uvs_params_ = np.average(uv_component_params, axis=0)
        uvs_ = UvSpectrum(uvs_params_)
        uv_bp = uvs_.eval(wv)
        uv_p_list += [uv_bp]

    UvP = np.array(uv_p_list).T
    UvD = UvP @ UvC

    if debug:
        from matplotlib.gridspec import GridSpec
        from MatrixData import simple_plot_3d
        plt.push()
        fig = plt.figure(figsize=(21,11))
        gs = GridSpec(2,3)

        ax1 = fig.add_subplot(gs[0,0], projection='3d')
        ax2 = fig.add_subplot(gs[0,1])
        ax3 = fig.add_subplot(gs[0,2])
        ax3.set_yscale('log')
        ax4 = fig.add_subplot(gs[1,0], projection='3d')
        ax5 = fig.add_subplot(gs[1,1])
        ax6 = fig.add_subplot(gs[1,2])

        fig.suptitle("Model Data with Linear Baseline Drift", fontsize=30)
        ax1.set_title("Xray Apparant 3D View", fontsize=20)
        ax2.set_title("Xray Apparant Elution with Components", fontsize=20)
        ax3.set_title("Xray Component Scattering Curves", fontsize=20)
        ax4.set_title("UV Apparant 3D View", fontsize=20)
        ax5.set_title("UV Apparant Elution with Components", fontsize=20)
        ax6.set_title("UV Component Spectrum Curves", fontsize=20)

        simple_plot_3d(ax1, D, x=qv)

        ax2.plot(x, by + y, '-', color='orange')
        stop = None if drift_type is None else -1
        for cy in cy_list[0:stop]:
            ax2.plot(x, by + cy, ':')

        if drift_type is not None:
            ax2.plot(x, by, ':', color='red')

        for p in p_list[0:stop]:
            ax3.plot(qv, p, ':')

        if drift_type is not None:
            ax3.plot(qv, bp, ':', color='red')

        simple_plot_3d(ax4, UvD, x=wv)

        ax5.plot(uv_x, uv_by + uv_y, '-', color='blue')
        for cy in uv_cy_list[0:stop]:
            ax5.plot(uv_x, uv_by + cy, ':')

        if drift_type is not None:
            ax5.plot(uv_x, uv_by, ':', color='red')

        for uv_p in uv_p_list[0:stop]:
            ax6.plot(wv, uv_p, ':')

        if drift_type is not None:
            ax6.plot(wv, uv_bp, ':', color='red')

        fig.tight_layout()
        fig.subplots_adjust(top=0.88)
        plt.show()
        plt.pop()

    ey = y + by
    uv_ey = uv_y + uv_by

    xr_curve = ElutionCurve(ey)
    rg_curve = RgCurveProxy(qv, xr_curve, cy_list)
    uv_curve = ElutionCurve(uv_ey)
    return (xr_curve, D), rg_curve, (uv_curve, UvD)

def extract_absorption_components(in_folder):
    from molass_legacy._MOLASS.SerialSettings import get_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from MatrixData import simple_plot_3d
    from molass_legacy.ElutionDecomposer import ElutionDecomposer

    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_corrected_sd()
    U, _, wv, uv_curve = sd.get_uv_data_separate_ly()

    print("U.shape=", U.shape, "wv.shape=", wv.shape)

    x = uv_curve.x
    y = uv_curve.y
    decomposer = ElutionDecomposer(uv_curve, x, y, deeply=True)
    opt_recs = decomposer.fit_recs

    cy_list = []
    for k, rec in enumerate(opt_recs):
        if k not in [1, 2, 4]:
            continue
        kno = rec.kno
        f = rec.evaluator
        cy = f(x)
        cy_list.append(cy)

    C = np.array(cy_list)
    P = U @ np.linalg.pinv(C)
    spectra_list = []
    for p in P.T:
        uvs = UvSpectrum()
        uvs.fit(wv, p, debug=False)
        print("uvs.params=", uvs.params)
        spectra_list.append(uvs)

    plt.push()
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))
    ax1.plot(x, y, color='gray', alpha=0.3)
    for k, cy in enumerate(cy_list):
        ax1.plot(x, cy, ':', label="component-%d" % k)
    ax1.legend()

    for k, uvs in enumerate(spectra_list):
        ax2.plot(wv, uvs.eval(wv), label="component-%d" % k)
    ax2.legend()

    fig.tight_layout()
    plt.show()
    plt.pop()



"""
    UV.FoldednessDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from  MatrixData import simple_plot_3d
from .Foldedness import Foldedness

def reconstruct_optimizer(dsets, optimizer):
    n_components = optimizer.n_components
    qvector = optimizer.qvector
    wvector = optimizer.wvector
    uv_base_curve = optimizer.uv_base_curve
    xr_base_curve = optimizer.xr_base_curve

    new_optimizer = optimizer.__class__(dsets, n_components,
                        qvector=qvector,
                        wvector=wvector,
                        uv_base_curve=uv_base_curve,
                        xr_base_curve=xr_base_curve,
                        debug=True,
                        )

    init_params = optimizer.init_params
    new_optimizer.prepare_for_optimization(init_params)

    return new_optimizer

def demo(caller, show_result_only=True):
    print("demo")
    dsets = caller.optinit_info.dsets
    optimizer = reconstruct_optimizer(dsets, caller.fullopt)
    params = caller.canvas.get_current_params()
    split_params = optimizer.split_params_simple(params)
    lrf_info = optimizer.objective_func(params, return_lrf_info=True, lrf_debug=False)
    Pxr, Cxr, Puv, Cuv, mapped_UvD = lrf_info.matrices

    if not show_result_only:

        with plt.Dp():
            fig, (ax0, ax1, ax2, ax3) = plt.subplots(ncols=4, figsize=(20,5), subplot_kw=dict(projection="3d"))
            fig.suptitle("3D View of mapped_UvD")
            ax0.set_title("uvD")
            simple_plot_3d(ax0, optimizer.uvD)
            ax1.set_title("uvD_")
            simple_plot_3d(ax1, optimizer.uvD_)
            ax2.set_title("xrD")
            simple_plot_3d(ax2, optimizer.xrD)
            ax3.set_title("mapped_UvD")
            simple_plot_3d(ax3, mapped_UvD)

            fig.tight_layout()
            plt.show()

        def plot_conc_matrix(ax, C):
            cy_list = []
            for k, cy in enumerate(C):
                cy_list.append(cy)
                ax.plot(cy, ":")
            ty = np.sum(cy_list, axis=0)
            ax.plot(ty, ":", color="red")

        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("conc_matrix")
            plot_conc_matrix(ax1, Cuv)
            plot_conc_matrix(ax2, Cxr)
            fig.tight_layout()
            plt.show()

    uvD = optimizer.uvD
    wv = optimizer.wvector

    foldedness = Foldedness(wv)

    assert Cxr.shape[1] == Cuv.shape[1]
    elen = Cxr.shape[1]
    a, b = split_params[3]
    jxr = np.arange(elen) + optimizer.xr_curve.x[0]
    juv = a*jxr + b

    for M in (Pxr, Cxr, Puv, Cuv, mapped_UvD):
        print(M.shape)

    i280 = bisect_right(wv, 280)

    def plot_components(title, ax, ecurve, j_, P, C):
        ax.set_title(title)
        ax.plot(ecurve.x, ecurve.y)
        cy_list = []
        for k, c in enumerate(C):
            cy = c*P[i280,k]
            cy_list.append(cy)
            ax.plot(j_, cy, ":")
        ty = np.sum(cy_list, axis=0)
        ax.plot(j_, ty, ":", color="red")

    def plot_foldedness(title, ax, wv, P):
        ax.set_title(title)
        fy_list = []
        for k, p in enumerate(P.T[0:-1]):
            f = foldedness.compute(p)
            ax.plot(wv, p, ":", label="component-%d, Foldedness=%.1f" % (k+1, f))
        ax.legend()

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12,10))
        fig.suptitle("280, 275, 258 debug plot")
        plot_components("UV Decomposition", axes[0,0], optimizer.uv_curve, juv, Puv, Cuv)
        plot_components("XR Decomposition", axes[0,1], optimizer.xr_curve, jxr, Pxr, Cxr)
        plot_foldedness("UV Foldedness", axes[1,0], wv, Puv)
        axes[1,1].set_axis_off()
        fig.tight_layout()
        plt.show()



"""
    LjEghParams.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.BasicUtils import Struct
from .EghParams import EghParamsBase
from .SimpleSecParams import initial_guess
from .BaselineParams import get_num_baseparams
from molass_legacy._MOLASS.SerialSettings import get_setting
from SecTheory.SecEstimator import NUM_SEC_PARAMS
from SecTheory.LanJorgensonEgh import convert_to_xr_params_lj

RG_UPPER_BOUND = get_setting("RG_UPPER_BOUND")

class LjEghParams(EghParamsBase):
    def __init__(self, n_components):
        self.n_components = n_components
        self.num_baseparams = get_num_baseparams()
        self.integral_baseline = self.num_baseparams == 3
        self.use_K = False
        self.advanced = True
        self.poresize = None        # to be removed
        self.poreexponent = None    # to be removed

        nc = n_components - 1
        self.pos = []
        self.pos.append(0)      # [0] xr_params
        sep = 3*nc
        self.pos.append(sep)    # [1] xr_baseparams
        sep += self.num_baseparams
        self.pos.append(sep)    # [2] rgs
        sep += nc
        self.pos.append(sep)    # [3] mapping
        sep = sep+2
        self.pos.append(sep)    # [4] uv_params
        sep += nc
        self.pos.append(sep)    # [5] uv_baseparams
        sep += 5 + self.num_baseparams
        self.pos.append(sep)    # [6] mappable_range
        sep += 2
        self.pos.append(sep)    # [7] end

        self.num_params = sep   # note that this does not include sec params

    def __str__(self):
        return "LjEghParams(nc=%d)" % (self.n_components - 1)

    def get_estimator(self, editor, debug=True):
        if debug:
            from importlib import reload
            import Estimators.LjEghEstimator
            reload(Estimators.LjEghEstimator)
        from Estimators.LjEghEstimator import LjEghEstimator
        return LjEghEstimator(editor)

    def split_params(self, n, params):
        if self.num_params != len(params):
            raise ValueError("len(params)=%d != %d, which is calculated from number of components %d." % (len(params), self.num_params, n))

        ret_params = []
        for p, q in zip(self.pos[:-1], self.pos[1:]):
            ret_params.append(params[p:q])

        ret_params[0] = ret_params[0].reshape((n-1,3))  # xr_params

        return ret_params

    def split_params_simple(self, params, convert=True):
        req_params = params[0:-NUM_SEC_PARAMS]
        sec_params = params[-NUM_SEC_PARAMS:]
        Npc, tI = sec_params[[0,2]]
        split_params = self.split_params(self.n_components, req_params)
        if convert:
            xr_params = convert_to_xr_params_lj(split_params[0], tI, Npc)
        else:
            xr_params = split_params[0]

        self.separate_params = [xr_params] + split_params[1:] + [sec_params]
        return self.separate_params

    def split_as_unified_params(self, params, **kwargs):
        return self.split_params_simple(params, **kwargs)

    def compute_init_guess(self, xr_params):
        temp_guess = initial_guess(xr_params)     # uses only xr_params[:,1]
        Np = get_setting("numplates_pc")
        return Struct(  params=temp_guess.params,
                                                    #     Ti         Np
                        bounds=temp_guess.bounds + ((-2000, 0), (Np/1.5, Np*1.5)),  # (9600, 21600) for Np=14400
                        )

    def get_xr_param_bounds(self, temp_xr_params, Ti, Np):
        xr_h_max = np.max(temp_xr_params[:,0])
        m_ = np.average(temp_xr_params[:,1])
        s_ = (m_ - Ti)/np.sqrt(Np)
        m_allow = np.max(temp_xr_params[:,1])*0.2
        xr_bounds = []
        for h, m, t in temp_xr_params:      # temp_xr_params is not converted
            xr_bounds.append((0, xr_h_max*1.1))                     # h
            xr_bounds.append((max(0, m - m_allow), m + m_allow))    # m
            xr_bounds.append((s_*0.5, s_*1.5))                      # s
        return xr_bounds

    def get_extended_bounds(self, init_xr_params):
        bounds = super().get_extended_bounds(init_xr_params)
        return bounds

    def make_bounds_mask(self):
        bounds_mask = np.zeros(self.num_params + NUM_SEC_PARAMS, dtype=bool)
        nc = self.n_components - 1
        xr_base_start = 3*nc
        bounds_mask[0:xr_base_start] = np.array([True, True, False]*nc)      # 

        rg_start = xr_base_start + self.num_baseparams
        self.rg_start = rg_start

        if self.integral_baseline:
            bounds_mask[rg_start-1] = True      # xr baseline fouling
            bounds_mask[self.pos[6]-1] = True   # uv baseline fouling

        bounds_mask[rg_start:rg_start+nc] = True
        return bounds_mask

    def update_bounds_hook(self, masked_init_params):
        # this overrides  EghParamsBase.update_bounds_hook which modifies sigma bounds unnecessary for this class
        nc = self.n_components - 1
        rg_start = nc*2             # note that params here are masked
        for i in range(nc):
            j = rg_start + i
            masked_init_params[j] = min(RG_UPPER_BOUND, masked_init_params[j])

    def get_xr_parameter_names(self):
        nc = self.n_components - 1
        xr_names = []
        for k in range(nc):
            xr_names += ["$h_%d$" % k, r"$\mu_%d$" % k, r"$\sigma_%d$" % k]
        return xr_names

    def get_parameter_names(self):
        nc = self.n_components - 1

        xr_names = self.get_xr_parameter_names()
        xr_basenames = ["$xb_a$", "$xb_b$"]
        if self.num_baseparams == 3:
            xr_basenames += ["$xb_r$"]
        rg_names = ["$R_{g%d}$" % k for k in range(nc)]
        mapping_names = ["$mp_a$", "$mp_b$"]
        uv_names = ["$uh_%d$" % k for k in range(nc)]
        uv_basenames = ["$L$", "$x_0$", "$k$", "$b$", "$s_1$", "$s_2$", "$diffratio$"]
        if self.num_baseparams == 3:
            uv_basenames += ["$ub_r$"]
        mr_names = ["$mr_a$", "$mr_b$"]
        seccol_names = ["$N_{pc}$", "$r_p$", "$t_I$", "$t_0$", "$P$", "$m$"]    # task: unify with EghParams.get_common_parameter_names()

        return np.array(xr_names + xr_basenames + rg_names + mapping_names + uv_names + uv_basenames + mr_names + seccol_names)

    def get_rg_start_index(self):
        return self.pos[2]

    def get_mr_start_index(self):
        return self.pos[6]

    def get_peak_pos_array_list(self, x_array):
        n = self.n_components - 1
        pos_array_list = []
        for k in range(n):
            pos_array_list.append(x_array[:,3*k+1])     # mu in (h, mu, sigma)
        return pos_array_list

    def get_params_sheet(self, parent, params, dsets, optimizer):
        from .LjEghParamsSheet import LjEghParamsSheet
        return LjEghParamsSheet(parent, params, dsets, optimizer)

    def split_get_unified_sec_params(self, params):
        separate_parame = self.split_params_simple(params)
        t0, K, rp, m = separate_parame[-2]
        Ti, Np = separate_parame[-1]
        return t0, K, rp, m, None, None, None,None, Ti, Np, None



# coding: utf-8
"""
    Baseline.BpaInspect.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import copy
from bisect import bisect_right
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import animation
from matplotlib.patches import Polygon
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.DENSS.DenssUtils import fit_data
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy._MOLASS.SerialSettings import get_setting
from MatrixData import simple_plot_3d
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from LPM import LPM_3d
from DataUtils import cut_upper_folders
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import BlinkingFrame
from molass_legacy.Mapping.MapperConstructor import create_mapper
from DataUtils import get_in_folder
import molass_legacy.KekLib.DebugPlot as dplt

class DebugPlotter:
    def __init__(self, sd, indeces):
        self.sd = sd
        self.qv = sd.qvector
        self.indeces = indeces
        ncols = len(indeces)
        fig, axes = dplt.subplots(nrows=1, ncols=ncols, figsize=(4.4*ncols, 6))
        fig.suptitle("LPM Debug for " + get_in_folder(), fontsize=20)
        self.fig = fig
        self.axes = axes
        self.bases = []

    def plot(self, i, base):
        try:
            w = np.where(self.indeces == i)[0]
            if len(w) == 0:
                return
            k = w[0]
        except:
            return

        self.bases.append(base)
        ax = self.axes[k]
        ax.set_title("Q[%d]=%.4g" % (i, self.qv[i]), fontsize=16)
        x = self.sd.xray_curve.x
        y = self.sd.intensity_array[:,i,1]
        y_ = y + base
        ax.plot(x, y_, color='orange')
        ax.plot(x, base, color='red')

    def show(self):
        ax_lims = []
        for ax in self.axes:
            ax_lims.append(ax.get_ylim())
        ax_lims = np.array(ax_lims)
        ymin = np.min(ax_lims[:,0])
        ymax = np.max(ax_lims[:,1])
        for ax in self.axes:
            ax.set_ylim(ymin, ymax)
        self.fig.tight_layout()
        self.fig.subplots_adjust(top=0.85)
        dplt.show()

class BpaInspector(Dialog):
    def __init__(self, parent, sd):
        self.parent = parent
        self.sd1 = sd.get_copy()
        self.sd2 = sd
        self.qv = sd.qvector
        self.i1 = bisect_right(sd.qvector, 0.02)
        self.i2 = bisect_right(sd.qvector, 0.1)

        self.debug_obj = DebugPlotter(sd, np.arange(self.i2-2,self.i2+3))

        self.mapper = create_mapper(None, sd, full_width=True)  # full_width=True is required
        self.bp = self.sd2.apply_baseline_correction(self.mapper.get_mapped_info(), return_base=True, debug_obj=self.debug_obj)

        self.debug_obj.show()

        D1, E1, _, _ = self.sd1.get_xr_data_separate_ly()
        D2, E1, qv, ecurve = self.sd2.get_xr_data_separate_ly()
        self.D1 = D1
        self.D2 = D2

        self.x = np.arange(D1.shape[1])
        self.zeros = np.zeros(len(self.x ))
        self.axis_ylims = None

        self.in_folder = get_setting('in_folder')
        Dialog.__init__(self, parent, "BPA Inspector", visible=False)

    def show(self):
        self._show()

    def body(self, frame):
        cframe = Tk.Frame(frame)
        cframe.pack()
        bframe = Tk.Frame(frame)
        bframe.pack(fill=Tk.X)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)
        pframe = Tk.Frame(bframe)
        pframe.pack(side=Tk.RIGHT)

        self.fig = fig = plt.figure(figsize=(21, 11))
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        gs = GridSpec(2,3)
        ax00 = fig.add_subplot(gs[0,0], projection='3d')
        ax01 = fig.add_subplot(gs[0,1])
        ax02 = fig.add_subplot(gs[0,2])
        ax10 = fig.add_subplot(gs[1,0], projection='3d')
        ax11 = fig.add_subplot(gs[1,1])
        ax12 = fig.add_subplot(gs[1,2])
        self.axes = np.array([[ax00, ax01, ax02], [ax10, ax11, ax12]])

        fig.suptitle("Base Plane Inspection for " + get_in_folder(), fontsize=20)
        ax01.set_title("Q[%d]=%.4g" % (self.i1, self.qv[self.i1]), fontsize=16)
        ax02.set_title("Q[%d]=%.4g" % (self.i2, self.qv[self.i2]), fontsize=16)
        self.draw_3d(ax00, self.D1)
        self.draw_3d(ax10, self.D2)
        self.draw_ecurve_at(ax01, self.D1, self.i1, baseline=True)
        self.draw_ecurve_at(ax02, self.D1, self.i2, baseline=True, base=self.debug_obj.bases[2])
        self.draw_ecurve_at(ax11, self.D2, self.i1, zeroline=True)
        self.draw_ecurve_at(ax12, self.D2, self.i2, zeroline=True )

        fig.tight_layout()
        fig.subplots_adjust(top=0.9)
        self.mpl_canvas.draw()

    def buttonbox( self ):
        bottom_frame = Tk.Frame(self)
        bottom_frame.pack(fill=Tk.BOTH, expand=1)

        width = int(self.mpl_canvas_widget.cget('width'))
        padx = width*0.05

        tframe = Tk.Frame(bottom_frame)
        tframe.pack(side=Tk.LEFT, padx=padx)
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        space = Tk.Frame(bottom_frame, width=width*0.25)
        space.pack(side=Tk.RIGHT)

        box = Tk.Frame(bottom_frame)
        box.pack(side=Tk.RIGHT)

        self.same_scale = Tk.IntVar()
        self.same_scale.set(0)
        w = Tk.Checkbutton(box, text="same scale", variable=self.same_scale)
        w.pack(side=Tk.LEFT, padx=20, pady=5)
        self.same_scale.trace("w", self.same_scale_tracer)

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=20, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def draw_3d(self, ax, D):
        simple_plot_3d(ax, D, x=self.qv)

    def draw_ecurve_at(self, ax, D, i, baseline=False, zeroline=False, base=None):
        y = D[i,:]
        ax.plot(self.x, y, color='orange')

        if base is not None:
            ax.plot(self.x, base, color='pink', label='LPM base')

        if baseline:
            by = self.bp[i,:]
            ax.plot(self.x, by, color='red', label='BPA base')

        if zeroline:
            ax.plot(self.x, self.zeros, ':', color='red')

        if base is not None:
            ax.legend()

    def same_scale_tracer(self, *args):
        if self.axis_ylims is None:
            ylims = []
            for i in range(2):
                row = []
                for j in range(3):
                    ax = self.axes[i,j]
                    row.append(ax.get_ylim())
                ylims.append(row)
            self.axis_ylims = np.array(ylims)

        same_scale = self.same_scale.get()
        for i in range(2):
            ax = self.axes[i,2]
            j = 1 if same_scale else 2
            ylim = self.axis_ylims[i,j]
            ax.set_ylim(ylim)

        self.mpl_canvas.draw()
